import numpy as np
# import scipy.sparse as sp
from scipy.linalg import null_space, qr
from numba import njit

from ..model import symmetries as sym

# Original Matlab code:
# Copyright 2019 Jeremy Mason
#
# Licensed under the Apache License, Version 2.0, <LICENSE-APACHE or
# http://apache.org/licenses/LICENSE-2.0> or the MIT license <LICENSE-MIT or
# http://opensource.org/licenses/MIT>, at your option. This file may not be
# copied, modified, or distributed except according to those terms.
#
# Rewritten in python by Moritz Frewein 2025

def check_matlab_matrix(M, tol):
    data = np.genfromtxt('/Users/moritz/Documents/code/textom/textom/ressources/symmetrizedHSH/output/matrix.csv',
                         delimiter=',',dtype=np.complex128)
    print(np.allclose(data,M, atol=tol))

def HSH_symmetrization_matrix(n_max, symmetry):
    gen = sym.generators(symmetry)
    gen_sample = sym.generators('1')
    X = collect_symm(gen,gen_sample, n_max=n_max, 
                    precision=1e-12
                    # precision=1e-15
                    )
    return X

def collect_symm(gc, gs, n_max, precision, skip_ghosts=True):
    # collect_symm find sets of coefficients up to n_max that make the
    # hyperspherical harmonic expansion invariant to the cubic crystal point
    # group generators. These can  be interpreted as defining a compact basis
    # for an orientation distribution obeying the required symmetry, namely,
    # the symmetrized hyperspherical harmonics.
    # 
    # Inputs:
    #   gc, gs - crystal and sample symmetry generators (sample not allowed for textom)
    #   n_max - maximum value of N for which sets of coefficients are found.
    # 
    # Outputs:
    #   symm  - a cell array containing three columns which are the outputs X,
    #           L and M of the make_symm_pq function. The value of N starts at
    #           zero and increases with the row.

    # symm = np.empty((n_max + 1, 3), dtype=object)
    symm = {'0': np.array([[1.]], np.complex128)}

    if skip_ghosts:
        step=4
    else:
        step=2

    n = step
    while n <= n_max:
        X, L, M = make_symm(gc, gs, n, precision)
        if X is not None and np.size(X) > 0:
            # symm[n, 0] = X
            # symm[n, 1] = L
            # symm[n, 2] = M
            symm[str(n)] = X.T
        n += step

    return symm

def make_symm(gc, gs, n, TOL):
    # make_symm finds sets of coefficients that make the hyperspherical
    # harmonic expansion invariant to the specified point symmetry group
    # generators. These can then be interpreted as defining a compact basis for an
    # orientation distribution obeying the required symmetry, namely, the
    # symmetrized hyperspherical harmonics.
    # 
    # The difference with make_symm is in the way the eigenspaces are combined. The
    # more formal procedure followed here constructs projection matrices and finds
    # the subspace belonging to the intersection. This is slower, but has the
    # advantage of being supported by well-established theorems.
    # 
    # Inputs:
    #   n   - upper index of the hyperspherical harmonics. The calculation is
    #         formally independent for each value of n.
    #   lx  - rotation angle and spherical angles of the rotation axis for one of 
    #         the generators left multiplying the argument of the hyperspherical
    #         harmonics. That is, a symmetry of the sample. The angles [w, th, ph]
    #         should be in the usual intervals. x can be a, b, c or d.
    #   rx  - rotation angle and spherical angles of the rotation axis for one of 
    #         the generators right multiplying the argument of the hyperspherical
    #         harmonics. That is, a symmetry of the crystal. The angles [w, th, ph]
    #         should be in the usual intervals. x can be a, b, c or d.
    #   TOL - tolerance for the desired eigenvalue of a matrix. For all practical
    #         cases the desired eigenvalues are well-separated from the rest of the
    #         spectrum, so this does not need to be excessively small.
    #
    # Outputs:
    #   X   - matrix with orthonormal columns that forms a basis for the space of 
    #         expansion coefficients satisfying the requested symmetries. That is,
    #         a column gives the expansion coefficients required to define one of
    #         the symmetrized hyperspherical harmonics. The expansion uses complex
    #         hyperspherical harmonics and orders the coefficients
    #         lexicographically in (l, m).
    #   L   - values of l for the rows of X.
    #   M   - values of m for the rows of X.

    # store objects (since X, L, M may have different shapes)
    l = n // 2

    # symmetry generators
    la = gs[0, :]
    lb = gs[1, :]
    ra = np.array([0., 0., 0.])
    rb = np.array([0., 0., 0.])
    lc = np.array([0., 0., 0.])
    ld = np.array([0., 0., 0.])

    # crystal
    rc = gc[0, :]
    rd = gc[1, :]

    # The equation for the irreps of SO(4) below allows the Clebsch-Gordan
    # coefficients to be collected into a unitary matrix. Performing a similarity
    # transformation with this matrix has the effect of converting from the 
    # uncoupled to the coupled basis.

    dim = (n + 1) ** 2
    CG = np.zeros((dim, dim))

    j_vals = np.arange(0, n + 1)
    for a in range(n + 1):
        j = j_vals[a]
        m_vals = np.arange(-j, j + 1)
        for b, m in enumerate(m_vals):
            row = (j + 1) ** 2 - j + m
            C, m1, m2 = CleGor(l, l, j, m)
            for c in range(len(C)):
                col = (int(m1[c]) + l) * (2 * l + 1) + (int(m2[c]) + l) + 1
                CG[row - 1, col - 1] = C[c]   # MATLAB → Python index shift

    # CG = CG.tocsr()
    CR = complex_real(n)

    X = np.eye(dim) #speye(dim, format="csr")
    P = X @ X.T

    # Construct simultaneous eigenvectors of eigenvalue one of the irreps of SO(4).
    # Observe that the eigenspace of eigenvalue one of a matrix A is equivalent to
    # the right nullspace of the matrix (A - I). The irrep of SO(4) is constructed
    # using the following equation (a transform of Eq. 2 of the manuscript);
    # 
    # R^{a a}_{c \gamma d \delta}(g_l, g_r) = \sum_{\alpha' \beta' \alpha \beta}
    # C^{c \gamma}_{a \alpha' a \beta'} U^a_{\alpha' \alpha}(g_r^{-1})
    # U^a_{\beta' \beta}(g_l) C^{d \delta}_{a \alpha b \beta}
    # 
    # That is, the desired irrep of SO(4) is given by a similarity transformation
    # of the Kroneker product of irreps of SO(3). The effect of right multiplying
    # a row vector of hyperspherical harmonics with this irrep is to carry every
    # rotation g \rightarrow g_l g g_r. Since the eigenspectrum of a matrix is
    # unchanged by a unitary transformation, the calculation can be performed in
    # the uncoupled basis and transformed to the coupled basis afterwards.

    # def apply_rotation(axis, ref):
    #     Rot = R(2*l, axis, ref)
    #     # Ul = rotation_mat(l, axis[0], axis[1], axis[2])
    #     # Ur = rotation_mat(l, -ref[0], ref[1], ref[2])
    #     # R = np.kron(Ur, Ul)
    #     Y = sp_orth(null_space(Rot - np.eye(Rot.shape[0]),TOL).T)
    #     # Y = sp_orth(sp_null(Rot - np.eye(Rot.shape[0]), 1, TOL))
    #     Q = Y @ Y.T
    #     return Q

    # # apply generators
    # for axis, ref in [(la, ra), (lb, rb), (lc, rc), (ld, rd)]:
    #     Q = apply_rotation(axis, ref)
        
    #     # X = sp_orth(sp_null(P @ Q - np.eye(Q.shape[0]), 1, TOL))
    #     P = X @ X.T
    rou = int(-np.log10(TOL))

    # if np.any(gs[:,0]<2*np.pi):
    Ul = U(l, la[0], la[1], la[2])
    Ur = U(l, -ra[0], ra[1], ra[2])
    Ra = np.round( np.kron(Ur, Ul), rou)
    Y = orth(null_space((Ra - np.eye(Ra.shape[0])).T,TOL))
    Q = Y @ Y.conj().T
    X = orth(null_space(np.round((P @ Q  - np.eye(Q.shape[0])).T,rou),TOL))
    P = X @ X.conj().T

    Ul = U(l, lb[0], lb[1], lb[2])
    Ur = U(l, -rb[0], rb[1], rb[2])
    Rb = np.round( np.kron(Ur, Ul), rou)
    Y = orth(null_space((Rb - np.eye(Rb.shape[0])).T,TOL))
    Q = Y @ Y.conj().T
    X = orth(null_space(np.round((P @ Q  - np.eye(Q.shape[0])).T,rou),TOL))
    P = X @ X.conj().T

    Ul = U(l, lc[0], lc[1], lc[2])
    Ur = U(l, -rc[0], rc[1], rc[2])
    Rc = np.round( np.kron(Ur, Ul), rou)
    Y = orth(null_space((Rc - np.eye(Rc.shape[0])).T,TOL))
    Q = Y @ Y.conj().T
    X = orth(null_space(np.round((P @ Q  - np.eye(Q.shape[0])).T,rou),TOL))
    P = X @ X.conj().T

    Ul = U(l, ld[0], ld[1], ld[2])
    Ur = U(l, -rd[0], rd[1], rd[2])
    Rd = np.round( np.kron(Ur, Ul), rou)
    Y = orth(null_space((Rd - np.eye(Rd.shape[0])).T,TOL))
    Q = Y @ Y.conj().T
    X = null_space(np.round((P @ Q  - np.eye(Q.shape[0])).T,rou),TOL)

    # convert to coupled basis
    X = CG @ X
    X = np.array(CR @ clean(CR.conj().T @ np.round(X,rou), TOL))

    # numerical error checks
    def max_err(R):
        if X is None or X.size == 0:
            return 0.0
        return np.max(np.abs((CG @ R @ CG.T - np.eye(R.shape[0])) @ X))
    err_a = max_err(Ra)
    err_b = max_err(Rb)
    err_c = max_err(Rc)
    err_d = max_err(Rd)

    if X.size > 0:
        if n//2%2:
            print(f"n = {n}, sHSHs found: {X.shape[1]} (ghosts), numerical error: {max([err_a, err_b, err_c, err_d]):.3e}")
        else:
            print(f"n = {n}, sHSHs found: {X.shape[1]}, numerical error: {max([err_a, err_b, err_c, err_d]):.3e}")

    else:
        print(f"n = {n}, no sHSHs.")

    # Assign indices
    L = np.zeros(dim, dtype=int)
    M = np.zeros(dim, dtype=int)
    for l_val in range(n + 1):
        for m in range(-l_val, l_val + 1):
            ind = (l_val + 1) ** 2 - l_val + m
            L[ind - 1] = l_val  # index shift
            M[ind - 1] = m

    return X, L, M

@njit
def R( n, gl, gr=np.array([0.,0.,0.]) ):
    """Function that computes a HSH rotation matrix for a given order n

    Source: Mason, J. K., Acta Crystallographica Section A Foundations of 
        Crystallography 65, no. 4 (July 1, 2009): 259?66. 
        https://doi.org/10.1107/S0108767309009921.
        equation (10)

    Parameters
    ----------
    n : int/float
        order of the HSH
    gl, gr : 2D ndarray, float
        axis-angle rotations given by angles: [omega, theta, phi]
        dimensions: 0: rotation, 1: ome,tta,phis
        
    Return values
    ------------
    R : 2D ndarray, complex
        matrix to convert a vector of HSH coefficients to a different
        vector that results in the same ODF but rotated
    """
    Ul = U( n/2, gl[0],gl[1],gl[2])
    Ur_inv = U( n/2, gr[0], gr[1], gr[2], True)
    UrUl = np.kron( Ur_inv, Ul ).astype(np.complex128)
    CG = CGn( n ).astype(np.complex128)
    R = ( CG @ UrUl @ np.conj(CG.T) )
    return R

@njit
def U(j,ome,tta,phi,invert=False):
    # U converts a rotation angle and spherical angles of the rotation
    # axis into the equivalent irrep of SO(3).
    # 
    # Inputs:
    #   j  - specifies dimension of the irrep.
    #   ome  - rotation angle in the interval [0, 2 \pi].
    #   tta - polar angle of rotation axis in the interval [0, \pi].
    #   phi - aximuthal angle of rotation axis in the interval [0, 2 \pi].
    #
    # Outputs:
    #   U  - (2 j + 1)-dimensional representation of SO(3), using the conventions
    #        established in Eq. 6 on page 81 of D. A. Varshalovich et al, Quantum
    #        Theory of Angular Momentum, 1988. Rows and columns ordered in
    #        increasing values of m' and m.

    if invert:
        ome = - ome

    tmp = np.tan(ome / 2) * np.cos(tta)
    tmp = (1 - 1j * tmp) / np.sqrt(1 + tmp**2)
    r_base = 1j * np.exp(-1j * phi) * tmp
    c_base = -1j * np.exp(1j * phi) * tmp
    
    # Require w to be in [-\pi, \pi]
    w = np.mod(ome + np.pi, 2 * np.pi) - np.pi
    xi = 2. * np.arcsin(np.sin(w / 2.) * np.sin(tta))
    U = wigner_d(j, xi).astype(np.complex128)
    
    m = np.arange(-j,j+1)
    n = int( 2 * j + 1 )
    for a in range(n):
        U[a, :] = U[a, :] * r_base**m[a]
        U[:, a] = U[:, a] * c_base**m[a]

    # tmp = tan(w / 2.) * cos(th);
    # tmp = (1. - 1i * tmp) / realsqrt(1 + tmp^2.);
    # r_base = 1i * exp(-1i * ph) * tmp;
    # c_base = -1i * exp(1i * ph) * tmp;
    
    # % Require w to be in [-\pi, \pi]
    # w = mod(w + pi, 2. * pi) - pi;
    # xi = 2. * asin(sin(w / 2.) * sin(th));
    # U = wigner_little_d(j, xi);
    
    # m = -j:j;
    # n = 2 * j + 1;
    # for a = 1:n
    #     U(a, :) = U(a, :) * r_base^m(a);
    #     U(:, a) = U(:, a) * c_base^m(a);
    return U

@njit
def CGn(n):
    """ Calculates all Clebsch-Gordan coefficients for an order n
    and assigns them into a matrix for further handling
    """
    l = n/2
    CG = np.zeros( ((n + 1)**2, (n + 1)**2), np.float64)
    for a in np.arange( 0, n+1, 1, np.int64 ):
        for b in np.arange( -a, a+1, 1, np.int64 ):
            row = np.array( (a + 1)**2 - a + b -1 , np.int64 )
            C, m1, m2 = CleGor( l, l, a, b )
            for c in np.arange( 0, len(C), 1, np.int64 ):
                col = int( (m1[c] + l) * (2 * l + 1) + (m2[c] + l) )
                CG[row, col] = C[c,0]
    return CG

def rotation_mat(j, w, th, ph):
    # rotation_mat converts a rotation angle and spherical angles of the rotation
    # axis into the equivalent irrep of SO(3).
    # 
    # Inputs:
    #   j  - specifies dimension of the irrep.
    #   w  - rotation angle in the interval [0, 2 \pi].
    #   th - polar angle of rotation axis in the interval [0, \pi].
    #   ph - aximuthal angle of rotation axis in the interval [0, 2 \pi].
    #
    # Outputs:
    #   U  - (2 j + 1)-dimensional representation of SO(3), using the conventions
    #        established in Eq. 6 on page 81 of D. A. Varshalovich et al, Quantum
    #        Theory of Angular Momentum, 1988. Rows and columns ordered in
    #        increasing values of m' and m.

    tmp = np.tan(w / 2.0) * np.cos(th)
    tmp = (1.0 - 1j * tmp) / np.sqrt(1.0 + tmp**2)
    r_base = 1j * np.exp(-1j * ph) * tmp
    c_base = -1j * np.exp(1j * ph) * tmp

    # force w into [-pi, pi]
    w = np.mod(w + np.pi, 2.0 * np.pi) - np.pi
    xi = 2.0 * np.arcsin(np.sin(w / 2.0) * np.sin(th))

    U = wigner_d(j, xi).astype(np.complex128)  # expected shape (2j+1, 2j+1)

    m = np.arange(-j, j + 1)
    n = 2 * j + 1

    # apply phase factors
    for a in range(n):
        U[a, :] *= r_base**m[a]
        U[:, a] *= c_base**m[a]

    return U

# def sp_null(A, opt=0, TOL=1e-12):
#     """
#     Find null spaces of sparse or dense matrix A.
    
#     Parameters
#     ----------
#     A : array_like or sparse matrix
#         Input matrix.
#     opt : int
#         0 = right null space, 1 = left null space
#     TOL : float
#         Tolerance for singular values.

#     Returns
#     -------
#     N : ndarray
#         Orthonormal basis for the chosen null space.
#     """
#     if opt == 0:
#         N = null_space(A, rcond=TOL)
#     elif opt == 1:
#         N = null_space(A.T, rcond=TOL).T
#     else:
#         raise ValueError("Invalid option. Use 0 for right null space, 1 for left null space.")

#     # zero small values for consistency with MATLAB's cleanup
#     N[np.abs(N) < TOL] = 0.0
#     return N

# # import numpy as np
# # import scipy.sparse as sp
# # import scipy.sparse.linalg as spla
# from scipy.linalg import lu

# def sp_null(A, opt, TOL=1e-12):
#     """
#     Find left or right null space of a sparse matrix A.

#     Parameters
#     ----------
#     A : (m, n) sparse or dense array
#         Input matrix.
#     opt : int
#         0 -> left null space, 1 -> right null space.
#     TOL : float
#         Threshold for considering pivots as significant.

#     Returns
#     -------
#     N : ndarray
#         Left or right null space basis (not orthonormal).
#     """
#     if not sp.isspmatrix(A):
#         A = sp.csr_matrix(A)

#     if opt == 0:
#         # Zero-out tiny entries
#         A = A.tocoo()
#         mask = np.abs(A.data) >= TOL
#         A = sp.coo_matrix((A.data[mask], (A.row[mask], A.col[mask])), shape=A.shape).tocsr()

#         L, U, Q = luq(A, TOL)

#         diagU = np.abs(U.diagonal())
#         r = np.count_nonzero(diagU > TOL)

#         inv_L = np.linalg.inv(L.toarray())  # L is sparse but invert to dense
#         N = inv_L[r:, :]

#         # clean small values
#         N[np.abs(N) < TOL] = 0.0
#         return N

#     elif opt == 1:
#         return sp_null(A.T, 0, TOL).T
#     else:
#         raise ValueError("Invalid option. Use 0 (left) or 1 (right).")


# def luq(A, TOL=1e-12):
#     """
#     Perform LU-like factorization with permutation to construct
#     A = L * U * Q with block-triangular U.

#     Returns
#     -------
#     L : sparse matrix (m x m)
#     U : sparse matrix (m x n)
#     Q : sparse matrix (n x n)
#     """
#     if not sp.isspmatrix(A):
#         A = sp.csr_matrix(A)

#     m, n = A.shape

#     # Dense LU decomposition (SciPy's sparse lu does not expose P, Q in same way)
#     A_dense = A.toarray()
#     P, L, U = lu(A_dense)  # A = P @ L @ U

#     # Convert to A = L * U * Q form
#     if n < m:
#         p = m - n
#         L = np.hstack([L, np.vstack([np.zeros((n, p)), np.eye(p)])])
#         U = np.vstack([U, np.zeros((p, n))])

#     L = P.T @ L
#     Q = np.eye(n)

#     # Identify pivots
#     diagU = np.abs(np.diag(U))
#     p1 = np.where(diagU > TOL)[0]
#     p = len(p1)

#     # Permute rows and columns
#     r2 = np.setdiff1d(np.arange(m), p1)
#     r_perm = np.concatenate([p1, r2])
#     L = L[:, r_perm]
#     U = U[r_perm, :]

#     c2 = np.setdiff1d(np.arange(n), p1)
#     c_perm = np.concatenate([p1, c2])
#     U = U[:, c_perm]
#     Q = Q[c_perm, :]

#     # Update indices
#     p1 = np.arange(p)
#     r2 = np.arange(p, m)
#     c2 = np.arange(p, n)

#     # Eliminate U21 block
#     if p > 0 and r2.size > 0:
#         X = np.linalg.solve(U[p1, p1], U[p1, :][:, c2])
#         U[np.ix_(p1, c2)] = 0.0
#         Q[p1, :] = Q[p1, :] + X @ Q[c2, :]

#         X = np.linalg.solve(U[p1, p1], U[np.ix_(r2, p1)].T).T
#         L[:, p1] = L[:, p1] + L[:, r2] @ X
#         U[np.ix_(r2, p1)] = 0.0
#         U[np.ix_(r2, c2)] -= X @ U[np.ix_(p1, c2)]

#     # Clean up block structure
#     if r2.size > 0 and c2.size > 0:
#         r3 = p + np.where(np.max(np.abs(U[np.ix_(r2, c2)]), axis=1) > TOL)[0]
#         c3 = p + np.where(np.max(np.abs(U[np.ix_(r2, c2)]), axis=0) > TOL)[0]

#         if r3.size > 0 or c3.size > 0:
#             # Recursively decompose U22
#             Lr, Ur, Qr = luq(sp.csr_matrix(U[np.ix_(r3, c3)]), TOL)
#             U[np.ix_(r3, c3)] = Ur
#             L[:, r3] = L[:, r3] @ Lr
#             Q[c3, :] = Qr @ Q[c3, :]

#             q = np.count_nonzero(np.abs(np.diag(U)) > TOL)
#             U = np.block([
#                 [U[:q, :q], np.zeros((q, n - q))],
#                 [np.zeros((m - q, q)), np.zeros((m - q, n - q))]
#             ])
#         else:
#             U[np.ix_(r2, c2)] = 0.0

#     return sp.csr_matrix(L), sp.csr_matrix(U), sp.csr_matrix(Q)

def orth(A):
    """
    Compute an orthonormal basis for the columns of A.
    """
    Q,_,_ = qr(A, mode='economic', pivoting=True)  # 'economic' = reduced QR
    # Q,_ = qr(A, mode='economic')  # 'economic' = reduced QR
    return Q

def complex_real(n):
    # complex_real constructs a sparse matrix that performs the similarity
    # transformation to convert from the complex harmonics to the real
    # harmonics. Rows ordered by increasing l and by increasing m. Columns
    # ordered by increasing l, by increasing m, and by c before s.
    # 
    # Inputs:
    #   n - maximum value of l.
    # 
    # Outputs:
    #   U - sparse matrix that performs the transformation. 

    sqrt2 = np.sqrt(2.0)
    dim = (n + 1) ** 2
    # U = sp.lil_matrix((dim, dim), dtype=complex)
    U = np.zeros((dim, dim), np.complex128)

    for L in range(n + 1):
        row0 = (L + 1)**2 - L - 1  # MATLAB 1-based → Python 0-based
        col0 = (L + 1)**2 - 2*L - 1
        U[row0, col0] = 1j**L

        for M in range(1, L + 1):
            row_minus = row0 - M
            row_plus  = row0 + M
            col1 = col0 + 2*M - 1
            col2 = col0 + 2*M

            U[row_minus, col1] = (-1)**M * (1j)**L / sqrt2
            U[row_plus,  col1] = (1j)**L / sqrt2
            U[row_minus, col2] = (-1)**(M-1) * (1j)**(L-1) / sqrt2
            U[row_plus,  col2] = (1j)**(L-1) / sqrt2

    return U
    # return U.tocsr()

def clean_col(A, a, TOL):
    """
    Zero out small entries in column `a` of matrix A below tolerance TOL.

    % clean_col strips out any real and imaginary components of the entries in
    % A(:, a) with magnitudes below the threshold.

    % Inputs:
    %   A   - matrix containing the column.
    %   a   - column index.
    %   TOL - threshold below which a component is considered insignificant.
    %
    % Outputs:
    %   A   - matrix containing the column. 
    """
    col = A[:, a]
    col[np.abs(col.real) < TOL] = 0.0
    col[np.abs(col.imag) < TOL] = 0.0
    A[:, a] = col
    return A

def clean(A, TOL=1e-12):
    """
    Clean matrix A: normalize columns, remove tiny entries, 
    enforce orthonormality with forward/backward sweep.

    % clean attempts to construct an orthonormal column space for A with the 
    % minimum number of nonzero entries. Assumes that A has full column rank.

    % Inputs:
    %   A   - matrix specifying the column space.
    %   TOL - threshold below which an entry is considered insignificant.
    %
    % Outputs:
    %   A   - matrix whose columns form an orthonormal column space for A.
    """
    m, n = A.shape
    rou = int(-np.log10(TOL))

    # Initial processing
    for a in range(n):
        norm = np.linalg.norm(A[:, a])
        if norm > 0:
            A[:, a] /= norm
        # A = clean_col(A, a, TOL)

    # Forward sweep
    r1 = 0
    for a in range(n):
        # Find column containing most significant entry
        p = 0.0
        while p < TOL and r1 < m:
            r1 += 1
            slice_abs = np.abs(A[r1-1, a:n])
            p_idx = np.argmax(slice_abs)
            p = slice_abs[p_idx]
        b = a + p_idx

        # Should not happen
        if r1 >= m:
            print("A is not full column rank")
            break

        # Swap column into leading position (Swap column a with column b)
        # swap = A[:,a].copy()
        # A[:,a] = A[:,b]
        # A[:,b] = swap
        A[:, [a, b]] = A[:, [b, a]]

        # Make leading entry real
        lead = A[r1-1, a]
        if np.abs(lead) > 0:
            A[:, a] *= np.conj(lead) / np.abs(lead)

        # Cancel leading entries, breaks orthogonality and normality
        cols = np.where(np.abs(A[r1-1, a+1:n]) > TOL)[0] + (a+1)
        for col_idx in cols:
            factor = A[r1-1, col_idx] / A[r1-1, a]
            A[:, col_idx] -= factor * A[:, a]
            # A = clean_col(A, col_idx, TOL)
        # for col_idx in range(a+1, n):
        #     if np.abs(A[r1-1, col_idx]) > TOL:
        #         A[:, col_idx] -= (A[r1-1, col_idx] / lead) * A[:, a]
                # A = clean_col(A, col_idx, TOL)

    # Backward sweep to restore orthonormality
    # Q, _, _ = qr(np.fliplr(A), mode='economic', pivoting=True)
    Q, _ = qr(np.fliplr(A), mode='economic')
    A = np.round(np.fliplr(Q), rou) # this is my "clean_col"

    for a in range(n):
        # A = clean_col(A, a, TOL)
        # Make leading entry positive
        nz = np.flatnonzero(A[:, a])
        if len(nz) > 0:
            A[:, a] *= np.sign(A[nz[0], a])

    return np.round(A, rou)

@njit
def CleGor(j1, j2, J, M):
    # CleGor returns the Clebsch-Gordan coefficients for the specified
    # total angular momenta and coupled z angular momentum component. This form
    # is particularly convenient from a computational standpoint. Follows the
    # approach of W. Straub in viXra:1403.0263.
    # 
    # Inputs:
    #   j1 - first uncoupled total angular momentum.
    #   j2 - second uncoupled total angular momentum.
    #   j  - coupled total angular momentum.
    #   m  - coupled z angular momentum component.
    #
    # Outputs:
    #   C  - all of the nonzero Clebsch-Gordan coefficients for the specified
    #        inputs. Ordered in increasing values of m1.
    #   m1 - first uncoupled z angular momentum components.
    #   m2 - second uncoupled z angular momentum components.
    
    if (J < abs(j1 - j2) or J > j1 + j2 or abs(M) > J):
        # Nothing to do
        # But to the silly list for numba with a type
        ## THis is not working yet, just here to see if jitting helps
        print('asdf')
        # l = List.empty_list(int64)
        # return l, l, l
    else:
        m11 = (M - j1 - j2 + abs(j1 - j2 + M)) / 2
        m1n = (M + j1 + j2 - abs(j1 - j2 - M)) / 2
        
        m1 = np.arange(m11,m1n+1,1,np.float64)
        m2 = M - m1
        
        j_const = j1 * (j1 + 1) + j2 * (j2 + 1) - J * (J + 1)
        
        n = int(m1n - m11 + 1)
        # A is a tridiagonal symmetric matrix
        ## Switched from list to tuple here to enable numba typing
        A = np.zeros( (n, n), np.float64 )
        for a in range(n):
            A[a,a] = j_const + 2 * m1[a] * m2[a]
        for a in range(n-1):
            tmp = np.sqrt(j1 * (j1 + 1) - m1[a] * m1[a+1]) * np.sqrt(j2 * (j2 + 1) - m2[a] * m2[a+1])
            A[a, a + 1] = tmp
            A[a + 1, a] = tmp
        
        # A determines C up to sign and normalization
        ## Here, we have the problem that the la.null_space isn't numba supported, can we switch to 
        ##numpy.linalg.svd?
        # C = la.null_space(A)
        C = nullspace(A)
        C = np.sign(C[n-1]) * C / np.sqrt( np.conj(C.T) @ C )
        return C, m1, m2
    
@njit
def wigner_d(j,tta):
    # wigner_d constructs a Wigner little d matrix given a total angular 
    # momenum and an angle. This corresponds to the irrep of SO(3) for a rotation
    # about the y axis. Follows the approach of X. M. Feng et al in 
    # 10.1103/PhysRevE.92.043307.
    # 
    # Inputs:
    #   j     - specifies dimension of the matrix.
    #   tta   - rotation angle in the interval [0, 2 \pi].
    #
    # Outputs:
    #   d     - Wigner little d matrix, with the rows and columns ordered in
    #           increasing values of m' and m. 

    m = np.arange(-j,j+1)#np.arange(-j,j+1,1,np.int64)
    n = int(2*j+1)

    X = np.sqrt((j + m) * (j - m + 1)) / (2 * 1j)
    # Jy is a tridiagonal Hermitian matrix
    Jy = np.zeros((n, n),np.complex128)#np.zeros([n, n],dtype=complex)
    for a in range(1,n):
        b = n - a - 1
        Jy[a - 1, a] = -X[a]
        Jy[b + 1, b] =  X[a]

    # # Requires that eigenvectors be ordered with increasing eigenvalues
    w,v = np.linalg.eig(Jy)
    w_ord = np.argsort(np.real(w))
    V_temp = v[:,w_ord]
    ## I however need to change to complex here for V and W
    V = V_temp.astype(np.complex128)
    W = np.copy(V)
    for a in range(n):
        W[:, a] = W[:, a] * np.exp(-1j * m[a] * tta)

    d = W @ np.conj(V.T)
    return np.real(d)

@njit
def nullspace(A, atol=1e-13, rtol=1e-13):
    """Compute an approximate basis for the nullspace of A.

    The algorithm used by this function is based on the singular value
    decomposition of `A`.

    Source: https://scipy-cookbook.readthedocs.io/items/RankNullspace.html
        last accessed: 7 June 2023

    Parameters
    ----------
    A : ndarray
        A should be at most 2-D.  A 1-D array with length k will be treated
        as a 2-D with shape (1, k)
    atol : float
        The absolute tolerance for a zero singular value.  Singular values
        smaller than `atol` are considered to be zero.
    rtol : float
        The relative tolerance.  Singular values less than rtol*smax are
        considered to be zero, where smax is the largest singular value.

    If both `atol` and `rtol` are positive, the combined tolerance is the
    maximum of the two; that is::
        tol = max(atol, rtol * smax)
    Singular values smaller than `tol` are considered to be zero.

    Return values
    ------------
    ns : ndarray
        If `A` is an array with shape (m, k), then `ns` will be an array
        with shape (k, n), where n is the estimated dimension of the
        nullspace of `A`.  The columns of `ns` are a basis for the
        nullspace; each element in numpy.dot(A, ns) will be approximately
        zero.
    """

    A = np.atleast_2d(A)
    u, s, vh = np.linalg.svd(A)
    tol = max(atol, rtol * s[0])
    nnz = (s >= tol).sum()
    ns = vh[nnz:].conj().T
    return ns



import numpy as np
import sys
from time import time
from numba import njit, prange
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
from scipy import ndimage

from ..model import symmetries as sym
from ..model import rotation as rot

@njit#(parallel=True)
def find_clusters_numba(q, gen, threshold):
    """
    Find clusters in a 3D array where neighboring elements deviate by less than a given threshold.
    
    Parameters:
    - g: 4D numpy array with axis-angle orientations in the 4th dimesion.
    - ppgrp: str, proper point group.
    - threshold: Maximum deviation allowed between neighbors to consider them in the same cluster.
    
    Returns:
    - clusters: 3D array of the same shape as input with integer labels for each cluster.
    - num_clusters: The number of clusters found.
    """
    shape = q.shape[:3]
    # q = rot.QfromOTP( g.reshape((np.prod(shape), 3))  )
    # q = q.reshape( (*shape, 4) )

    # print('\tLooking for clusters')
    clusters = np.zeros(shape, np.int32)
    # visited = np.zeros(shape, np.bool_)
    visited = isnan_numba(q[:,:,:,0])
    vis_init = visited.sum()
    cluster_label = 0
    
    # Define neighbors for 3D connectivity (6-connected neighbors)
    neighbors = np.array([
        (1, 0, 0), (-1, 0, 0),
        (0, 1, 0), (0, -1, 0),
        (0, 0, 1), (0, 0, -1)
    ])

    # Preallocate stack with a maximum possible size (equal to the array size)
    max_stack_size = visited.size - vis_init
    stack = np.empty((max_stack_size, 3), dtype=np.int32)

    def grow_cluster(x, y, z, label):
        """Grow a cluster starting from (x, y, z) using the threshold and label."""
        stack_size = 0
        stack[stack_size] = np.array([x, y, z], dtype=np.int32)
        stack_size += 1
        clusters[x, y, z] = label
        visited[x, y, z] = True

        while stack_size > 0:
            cx, cy, cz = stack[stack_size - 1]
            stack_size -= 1
            # Check all 6 neighbors
            for dx, dy, dz in neighbors:
                nx, ny, nz = cx + dx, cy + dy, cz + dz
                # Check bounds
                if 0 <= nx < shape[0] and 0 <= ny < shape[1] and 0 <= nz < shape[2]:
                    if not visited[nx, ny, nz]:
                        # calculate misorientation
                        mori = rot.misorientation_single(
                            q[nx, ny, nz], q[cx, cy, cz],gen) 
                        if mori <= threshold:
                            # Mark as visited and add to cluster
                            visited[nx, ny, nz] = True
                            clusters[nx, ny, nz] = label
                            stack[stack_size] = np.array([nx, ny, nz], dtype=np.int32)
                            stack_size += 1

    # Iterate over each point in the 3D array in parallel
    for x in range(shape[0]):
        for y in range(shape[1]):
            for z in range(shape[2]):
                if not visited[x, y, z]:
                    # Start a new cluster with a unique label
                    cluster_label += 1
                    grow_cluster(x, y, z, cluster_label)

    return clusters, cluster_label

def plot_clusters_3d(clusters, min_size=100):
    """
    Plot clusters in 3D with each cluster in a different color.
    
    Parameters:
    - clusters: 3D numpy array with integer labels for each cluster.
    """
    # Get the unique cluster labels, excluding 0 (assuming 0 is the background or unlabeled)
    cluster_labels = np.unique(clusters)
    cluster_labels = cluster_labels[cluster_labels != 0]

    cluster_sizes = np.array([(clusters==n+1).sum() for n in cluster_labels])
    cluster_labels = cluster_labels[cluster_sizes > min_size]
    
    # Set up the colormap (use a colormap with enough unique colors for each cluster)
    cmap = cm.get_cmap("tab20", len(cluster_labels))  # e.g., "tab20" for categorical colors

    # Create a 3D plot
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    
    # Plot each cluster with a unique color
    for i, label in enumerate(cluster_labels):
        # Extract the x, y, z coordinates of points in the current cluster
        x, y, z = np.where(clusters == label)
        
        # Plot the cluster in 3D with a unique color from the colormap
        ax.scatter(x, y, z, color=cmap(i), label=f"Cluster {label}", s=5)
    
    # Add labels and a legend
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')
    plt.legend()
    plt.show()

def mori_3d( g, ppgrp ):
    """Calculates misorientations between neighboring voxels

    Parameters
    ----------
    g : ndarray
        orientations in axis-angle (3-angle) parametrization
    ppgrp : str
        proper point group

    Returns
    -------
    ndarray
        misorientation arrays
    """
    q = rot.QfromOTP( # make quaternions
        g.reshape((np.prod(g.shape[:3]), 3)) # flatten
        ).reshape((*g.shape[:3], 4)) # shape back
    
    gen = sym.generators(ppgrp)
    Qgen = rot.QfromOTP(gen)
    pgroup = rot.generate_group(Qgen)

    nx,ny,nz,_ = q.shape
    mori_x = rot.misorientation_angle_stack(
        q[:-1,:,:].reshape((nx-1)*ny*nz,4), 
        q[1:,:,:].reshape((nx-1)*ny*nz,4), 
        pgroup,#gen
    ).reshape((nx-1),ny,nz)
    mori_y = rot.misorientation_angle_stack(
        q[:,:-1,:].reshape(nx*(ny-1)*nz,4), 
        q[:,1:,:].reshape(nx*(ny-1)*nz,4), 
        pgroup,#gen
    ).reshape(nx,(ny-1),nz)
    mori_z = rot.misorientation_angle_stack(
        q[:,:,:-1].reshape(nx*ny*(nz-1),4), 
        q[:,:,1:].reshape(nx*ny*(nz-1),4), 
        pgroup,#gen
    ).reshape(nx,ny,(nz-1))

    mori_max = np.max(np.array((
        np.append(np.pi*np.ones((1,ny,nz)), mori_x, axis=0),
        np.append(mori_x,np.pi*np.ones((1,ny,nz)), axis=0),
        np.append(np.pi*np.ones((nx,1,nz)), mori_y, axis=1),
        np.append(mori_y, np.pi*np.ones((nx,1,nz)), axis=1),
        np.append(np.pi*np.ones((nx,ny,1)), mori_z, axis=2),
        np.append(mori_z, np.pi*np.ones((nx,ny,1)), axis=2),
    )), axis=0)

    return mori_x, mori_y, mori_z, mori_max

def label_zones( array_3d, tresh, min_size=None, max_zones=32 ):

    array_3d[np.isnan(array_3d)] = np.pi

    # Label the connected components (zones) below a treshold
    # labeled_zones, num_zones = label(binary_dilation(array_3d < tresh))
    labeled_zones, num_zones = ndimage.label(array_3d < tresh)

    # Calculate the size of each zone
    zone_sizes = np.bincount(labeled_zones.ravel())
    
    # Find zones that meet the minimum size requirement
    valid_zones = [(label, size) for label, size in enumerate(zone_sizes) if size >= min_size and label != 0]
    
    # Sort the valid zones by size in descending order and keep only the largest `max_zones`
    largest_zones = sorted(valid_zones, key=lambda x: x[1], reverse=True)[:max_zones]
    largest_zone_labels = [label for label, size in largest_zones]
    zone_sizes = np.array([size for label, size in largest_zones])

    # Create a new array for relabeling
    zones = np.zeros_like(labeled_zones)
    for new_label, old_label in enumerate(largest_zone_labels, start=1):
        zones[labeled_zones == old_label] = new_label  # Relabel sequentially

    return zones, zone_sizes


    # zone_sizes = []
    # for n in range(1, num_zones+1):
    #     size = (labeled_zones==n).sum()
    #     if size >= min_zone_size:
    #         labeled_zones[labeled_zones==n] = 0
    #         zone_sizes.append(size)
    # zidx = np.argsort(zone_sizes)
    # for k in range(zone_sizes.size - max_zone_no):
    #     labeled_zones[labeled_zones==np.where(zidx==k)[0]] = 0

    # return labeled_zones, np.array(zone_sizes)

@njit
def isnan_numba(array):
    # Create a boolean mask where True indicates NaN values
    mask = np.zeros(array.shape, dtype=np.bool_)
    for i in range(array.shape[0]):
        for j in range(array.shape[1]):
            for k in range(array.shape[2]):
                # Check if the value is NaN by comparing the value to itself
                if array[i, j, k] != array[i, j, k]:
                    mask[i, j, k] = True
    return mask




import numpy as np
import os
from time import time
import sys
from numba import njit
from resource import getrusage, RUSAGE_SELF
import matplotlib.pyplot as plt

from .fitting import fitting
from ..misc import timestring
from ...version import __version__

def optimize( fit:fitting, mode, alg='simple', 
            project_c0=True, project_all=False,
            step_size_0=1e-5, optimize_stepsize=True,
            tol=1e-3, minstep=1e-2, itermax=3000,
            name='', ):
    """ Gradient descent optimizer for texture tomography

    Parameters
    ----------
    fit : class object
        object from the fitting module containing loss function and gradient
    mode : int
        decides what to optimize
        0: only zero order, 1: only highest order, 2: all orders
    -- optional --
    name : str
        add sth to the name of the output file
    tol : float
        stops optimization if the relative change of the norm of the
        gradient deceeds this value
    minstep : float
        stops optimization if the initial stepsize is reduced by this factor
    itermas : int
        stops optimization if this number of iterations is reached
    
    Return values
    ------------
    c : 1D ndarray, float
        solution found by the optimization in terms of sHSH coefficients
    opt : dict
        collection of results:
        'c' : 1D ndarray, float
            solution found by the optimization in terms of sHSH coefficients
        'loss' : 1D ndarray, float
            value of the loss function at each iteration
        'nGrad' : 1D ndarray, float
            norm of the gradient at each iteration
        'proj' : 1D ndarray, bool
            marks if c0 was projected or not for each iteration
        'prec' : float
            final precision
        'fgam' : float
            final stepsize modification
        't_total' : float
            total optimization time in s
        't_iter_av' : float
            average time per iteration in s
        'max_memory' : float
            memory usage of the process
    """

    t0 = time()
    c = fit.C
    match mode:
        case 0: # fit c0 from average intensity
            print('Start optimizing average intensity (mode 0)')
            modename = 'c0'
            def fun( C ):
                return fit.loss_c0( C )
            def jac( C ):
                return fit.grad_c0( C )
        case 1: # fit only highest order
            print(f'Start optimizing order {fit.ns[-1]}, mode {mode}')
            modename = 'hc'
            def fun( C ):
                return fit.loss( C )
            def jac( C ):
                return fit.grad_highest( C )
        case 2: # fit all orders
            print(f'Start optimizing order {fit.ns[-1]}, mode {mode}')
            modename='C'
            def fun( C ):
                return fit.loss( C )
            def jac( C ):
                return fit.grad_full( C )
        case 3: # fit all orders but 0
            print(f'Start optimizing order {fit.ns[-1]}, mode {mode}')
            modename='nc0'
            def fun( C ):
                return fit.loss( C )
            def jac( C ):
                return fit.grad_allbut0( C )
        case 4: # experimental, bring all together
            print(f'Start optimizing all parameters (mode 4)')
            modename='magic'
            def fun( C ):
                return fit.loss( C )
            def jac( C ):
                return fit.grad_full( C )
            def jac_calib( C ):
                return fit.grad_allbut0( C )
            # do i need jac_calib? don't i just need to take grad[:,1:] ??
            
    # Initialize functions    
    loss = fun( c )
    grad = jac( c )
    t1 = time()
    print('\tCompiled functions, %.3f s' % ((t1-t0)) )
    
    # calculate inverse lipschitz constants
    L = fit.lipschitz().reshape((c.shape[0],1))
    L[ L < 1e-4*L.max() ] = 0.
    invL = np.zeros_like(L)
    invL[L!=0] = 2/L[L!=0]
    gam = np.tile( invL, (1,c.shape[1]) )
    direction = - gam * grad
    t2 = time()
    if optimize_stepsize:
        step_size, _ = double_line_search(fit.loss,fit.C,direction,step_size_0,
                                max_steps=100,tol=1e-6)

        t2 = time()
        if step_size == 0:
            print('\tOptimial stepsize not found, lower step_size_0 by a few orders of magnitude')
            return 0
        else:
            print(f'\tOptimized stepsize: {step_size:.3e}, {t2-t1:.3f} s' )

        if mode==4:
            grad_2 = jac_calib( c )
        
            # calculate inverse lipschitz constants
            L = fit.lipschitz().reshape((c.shape[0],1))
            L[ L < 1e-4*L.max() ] = 0.
            invL = np.zeros_like(L)
            invL[L!=0] = 2/L[L!=0]
            gam_2 = np.tile( invL, (1,c.shape[1]) )
            direction_2 = - gam_2 * grad_2
            step_size_2, _ = line_search(fit.loss,fit.C,direction_2,step_size,
                                    max_steps=100,tol=1e-6)
            t3 = time()
            if step_size == 0:
                print('\tOptimial stepsize not found, lower step_size_0 by a few orders of magnitude')
                return 0
            else:
                print(f'\tOptimized higher order stepsize: {step_size_2:.3e}, {t3-t2:.3f} s' )

            step_size = np.column_stack([
                step_size * np.ones_like(fit.C[:,0]),
                step_size_2 * np.ones_like(fit.C[:,1:])
            ])
            t2=t3
    else:
        step_size = step_size_0
    
    opt = {}
    opt['step_size'] = step_size

    # initialize iterables
    c_next = c
    fgam, prec = 1,1
    nGrad = _vnorm(grad)
    c0_proj = []
    opt['loss'] = [loss]
    # f,ax = plt.subplots()
    # line = ax.plot([0], opt['loss'])
    # ax.set_xlabel( 'Iterations' )
    # ax.set_ylabel( 'Loss function' )
    opt['nGrad'] = [nGrad]
    opt['proj'] = []
    opt['max_memory'] = getrusage(RUSAGE_SELF)
    iter = 0
    t_iter_av = 0.
    try:
        # start optimization
        while ( 
            ( prec > tol ) and 
            ( iter < itermax ) and
            ( fgam > minstep)
            ):        

            # specify search direction
            direction = - gam * grad

            # determinate stepsize according to algorithm
            if alg=='backtracking':
                step_size = backtracking_line_search(
                    fun, grad, c, direction, c1=1e-4, alpha0=step_size, beta=0.5, min_alpha=1e-6)
            elif alg=='simple':
                step_size = fgam * step_size
            elif alg=='quadratic':
                step_size = quadratic_interpolation_line_search(
                    fun, grad, c, direction, c1 = 1e-4, alpha0=step_size, max_iters=10)
            
            # Update parameters with the found step size
            c_next = c + step_size * direction
            
            # project 0-order coefficients to non-negative values
            if project_all:
                c_next, c0_proj = _project_negative_to_zero(c_next)
            elif project_c0:
                c_next[:,0], c0_proj = _project_negative_to_zero(np.atleast_2d(c_next[:,0]))

            loss_next = fun( c_next )
            if loss_next < loss: # check if loss function decreased
                iter += 1
                # accept new values
                c = c_next
                loss = loss_next
                # get new gradient
                grad = jac( c )
                # get precision of fit
                nGrad_m1 = nGrad
                nGrad = _vnorm(grad)
                prec = np.abs(nGrad_m1-nGrad)/nGrad
                # write into opt dictionary
                opt['loss'].append( loss )
                opt['nGrad'].append( nGrad )
                opt['proj'].append( c0_proj )
                opt['max_memory'] = max( getrusage(RUSAGE_SELF), opt['max_memory'] )
                
                # update_plot(f,ax,line,loss)

                t_iter_av = (time()-t2)/iter
                sys.stdout.write('\r\tIt %d, loss: %.7e, t/it: %.2f s, precision: %.2e       ' % (
                    iter,loss,t_iter_av, prec))
                sys.stdout.flush()
            else: 
                fgam *= 0.8
                # break
    finally:
        opt['c'] = c # optimized parameters
        opt['t_iter_av'] = t_iter_av # average time per iteration in s
        opt['t_total'] = time()-t0 # total optimization time in s
        opt['prec'] = prec # final precision
        opt['MDL'] = fit.MDL(loss)
        opt['flag_use_peak'] = fit.flag_use_peak
        opt['version'] = __version__
        # opt['fgam'] = fgam # final stepsize modification
        print('\n\tFinished. total time: %.2f min' % ( (time()-t0)/60 ))
        if fgam <= minstep:
            print('\t\tMinimum stepsize reached')

        # write solution to file
        if name != '':
            name = '_'+name
        resname = 'opt_%s_%s_%s%u%s_%s.h5' % (
            fit.title, modename, fit.odf_mode, fit.odf_par, name, timestring())
        opt['resname'] = resname
        # os.makedirs('output/', exist_ok=True)
        # with open('output/'+resname, 'wb') as file:
        #     dump(opt, file)
        return c, opt

def backtracking_line_search(
        loss, grad, theta, direction, 
        c1 = 1e-4, alpha0=1.0, beta=0.5, min_alpha=1e-2 ):
    """Rudimentary search for a stepsize that gives us sufficient progress as measured by the Armijo condition

    A common choice for alpha_0 is alpha_0 = 1, but this can vary
    somewhat depending on the algorithm. The choice of c1 can range
    from extremely small (10e-4, encouraging larger steps) to relatively
    large (0.3, encouraging smaller steps), and typical values of beta range
    from 0.1, (corresponding to a relatively coarse search) to 0.8 (corre-
    sponding to a finer search).
    https://mdav.ece.gatech.edu/ece-3803-fall2021/notes/11-notes-3803-f21.pdf

        Parameters:
        loss (function): Objective function to minimize.
        grad (np.array): Current gradient.
        theta (np.array): Current parameter vector.
        direction (np.array): Descent direction (e.g., negative gradient).
        c1 (float): Armijo condition constant for sufficient decrease.
        alpha_0 (float): Initial step size.
        min_alpha (float): Minimum step size.
        
    Returns:
        alpha (float): Step size that approximately minimizes the loss along `direction`.
    """
    # Evaluate the loss and gradient at the starting point
    f0 = loss(theta)
    g0 = np.einsum('ij,ij->',direction,grad) #this does not give the good gradient - why?

    step_size = alpha0
    # Check Armijo condition for sufficient decrease
    while loss(theta + step_size * direction) > f0 + c1 * step_size * g0:
        step_size *= beta  # Reduce step size by factor beta
        if step_size <= min_alpha:
            break
    return step_size

# @njit
# def line_search(f, x, p, alpha_init=1.0, max_steps=10, tol=1e-6):
#     """
#     Performs a line search to find the best step size along direction p from point x.

#     Parameters:
#     f          : function that takes a 1D array and returns a scalar
#     x          : current position (1D array)
#     p          : search direction (1D array)
#     alpha_init : initial step size (float)
#     max_steps  : maximum number of steps in each phase
#     tol        : convergence tolerance for step size

#     Returns:
#     best_alpha : optimal step size
#     best_f     : loss value at optimal step
#     """
#     alpha = alpha_init
#     f0 = f(x)
#     best_alpha = 0.0
#     best_f = f0

#     # Expand phase: try larger steps as long as it improves the loss
#     for _ in range(max_steps):
#         x_new = x + alpha * p
#         f_new = f(x_new)
#         if f_new < best_f:
#             best_alpha = alpha
#             best_f = f_new
#             alpha *= 2.0
#         else:
#             break

#     # Contract phase: binary search between best_alpha and last tried alpha
#     alpha_high = alpha
#     alpha_low = best_alpha

#     for _ in range(max_steps):
#         alpha = 0.5 * (alpha_low + alpha_high)
#         x_new = x + alpha * p
#         f_new = f(x_new)

#         if f_new < best_f:
#             best_alpha = alpha
#             best_f = f_new
#             alpha_low = alpha
#         else:
#             alpha_high = alpha

#         if abs(alpha_high - alpha_low) < tol:
#             break

#     return best_alpha, best_f

def line_search(function, coefficients, direction, alpha_init=1.0, max_steps=10, tol=1e-6):
    alpha = alpha_init
    f0 = function(coefficients)
    best_alpha = 0.0
    best_f = f0

    # Expand phase (increase alpha)
    for _ in range(max_steps):
        x_new = coefficients + alpha * direction
        f_new = function(x_new)
        if f_new < best_f:
            best_alpha = alpha
            best_f = f_new
            alpha *= 2.0
        else:
            break  # start contracting

    # Contract phase (binary search or geometric)
    alpha_high = alpha
    alpha_low = best_alpha
    for _ in range(max_steps):
        alpha = 0.5 * (alpha_low + alpha_high)
        x_new = coefficients + alpha * direction
        f_new = function(x_new)
        if f_new < best_f:
            best_alpha = alpha
            best_f = f_new
            alpha_low = alpha
        else:
            alpha_high = alpha

        if abs(alpha_high - alpha_low) < tol:
            break

    return best_alpha, best_f

def double_line_search(funciton, coefficients, direction, alpha_init=1.0, max_steps=10, tol=1e-6):
    alpha = alpha_init
    f0 = funciton(coefficients)
    best_alpha = 0.0
    best_f = f0

    def expand_contract(alpha, best_alpha, best_f, lam):
        # Expand phase (increase alpha)
        for _ in range(max_steps):
            x_new = coefficients + alpha * direction
            f_new = funciton(x_new)
            if f_new < best_f:
                best_alpha = alpha
                best_f = f_new
                alpha *= lam
            else:
                break  # start contracting

        # Contract phase (binary search or geometric)
        alpha_high = alpha
        alpha_low = best_alpha
        for _ in range(max_steps):
            alpha = 1/lam * (alpha_low + alpha_high)
            x_new = coefficients + alpha * direction
            f_new = funciton(x_new)
            if f_new < best_f:
                best_alpha = alpha
                best_f = f_new
                alpha_low = alpha
            else:
                alpha_high = alpha

            if abs(alpha_high - alpha_low) < tol:
                break
        return alpha, best_alpha, best_f
    
    # first phase, coarse search
    alpha, best_alpha, best_f = expand_contract(
        alpha, best_alpha, best_f, 10.)
    # second phase, fine search
    _, best_alpha, best_f = expand_contract(
        alpha, best_alpha, best_f, 2.)

    return best_alpha, best_f

def quadratic_interpolation_line_search(loss, grad, theta, direction, c1 = 1e-4, alpha0=1.0, max_iters=10):
    """Quadratic interpolation line search along the search direction.
    
    Parameters:
        loss (function): Objective function to minimize.
        grad (np.array): Current gradient.
        theta (np.array): Current parameter vector.
        direction (np.array): Descent direction (e.g., negative gradient).
        c1 (float): Armijo condition constant for sufficient decrease.
        alpha0 (float): Initial step size.
        max_iters (int): Maximum number of line search iterations.
        
    Returns:
        alpha (float): Step size that approximately minimizes the loss along `direction`.
    """
    
    # Evaluate the loss and gradient at the starting point
    f0 = loss(theta)
    g0 = np.einsum('ij,ij->',direction,grad)
    
    alpha_curr = alpha0  # Current step size guess
    f_curr = loss(theta + alpha_curr * direction)  # Loss at current alpha

    for _ in range(max_iters):
        # Check Armijo condition for sufficient decrease
        if f_curr <= f0 + c1 * alpha_curr * g0:
            return alpha_curr
        
        if alpha_curr == 0:
            break

        # Fit a quadratic model based on the two points: (0, f0) and (alpha_curr, f_curr)
        a = ((f_curr - f0) - alpha_curr * g0) / alpha_curr**2
        
        # Calculate the minimum of the quadratic model
        alpha_star = -g0 / (2 * a)
        
        # Clip alpha_star to be within reasonable bounds
        if alpha_star < 0 or alpha_star > alpha_curr / 2:
            alpha_star = alpha_curr / 2
        
        # Update interval for next iteration
        alpha_curr = alpha_star
        f_curr = loss(theta + alpha_curr * direction)

    return alpha_curr

def _vnorm( v ):
    " calculates the vector norm "
    if isinstance(v,np.ndarray):
        return (v**2).sum()**(1/2)
    else:
        return 1

@njit
def _project_negative_to_zero(c):
    "projects any negative values to 0"
    proj = False
    for i in range(c.shape[0]):
        for j in range(c.shape[1]):
            if c[i, j] < 0:
                c[i, j] = 0
                proj = True
    return c, proj

def update_plot(fig, axis,line_obj,loss):
    
    line_obj(range(len(loss)), loss)
    axis.relim()            # Recalculate limits
    axis.autoscale_view()   # Autoscale if needed
    fig.canvas.draw()
    fig.canvas.flush_events()  # Real-time update

    plt.pause(0.3)  # Pause to visualize iteration (optional)


import numpy as np
from numba import njit, prange
from orix.sampling import get_sample_fundamental
import orix.quaternion.symmetry as osym 
from orix.quaternion import Orientation, Rotation
from scipy.spatial import ConvexHull

from . import symmetries as sym
from .. import numba_plugins as nb
from ...config import data_type

def sample_fundamental_zone( dchi=2*np.pi/120, sampling='cubochoric', symmetry='222', ):
    """Defines the grid of angles for crystallite rotation

    Parameters
    ----------
    dchi : _type_, optional
        _description_, by default 2*np.pi/120
    sampling : str, optional
        _description_, by default 'cubochoric'
    symmetry : str, optional
        _description_, by default '222'

    Returns
    -------
    Gc : 2D ndarray, float
        register of axis-angle rotations
        dim: 0: rotation index, 1: [omega, theta, phi]
    dV : 1D ndarray, float
        volume element for integrating over the odf
        dim: 0: rotation index
    V_fz : float
        volume of the fundamental zone
    """
    try:
        if sampling=='cubochoric':
            pg = getattr( osym, sym.get_SFnotation( symmetry ) )
            rot_orix = get_sample_fundamental(
                    dchi*180/np.pi, 
                    point_group= pg,
                    method=sampling
            )
            
            Gc = OTPfromQ(rot_orix.data)

            # cubochoric is equal volume mapped
            # Sing and De Graef, 2016
            dV = 1/dchi**3 * np.ones( 
                Gc.shape[0], data_type)
    except:
        sampling = 'simple'
    
    if sampling == 'simple':
        ## set up angles used to rotate the single crystal patterns
        ome = np.linspace( 
            dchi/2, np.pi-dchi/2, int(np.pi/dchi), endpoint=True)
        tta = np.linspace( 
            dchi/2, np.pi-dchi/2, int(np.pi/dchi), endpoint=True)
        phi = np.linspace( 0, 2*np.pi, int(2*np.pi/dchi), endpoint=False)
        TTA, PHI, OME = np.meshgrid(tta, phi, ome)
        Ome, Tta, Phi = OME.flatten(), TTA.flatten(), PHI.flatten()
        Gc = np.column_stack((Ome,Tta,Phi))

        ## apply the crystal symmetry conditions on Tta, Phi, Ome, so in the following 
        # only these are calculated by only choosing rotations in the fundamental zone 
        # of the proper point group
        fz = sym.zone(symmetry,Gc)
        Tta, Phi, Ome = Tta[fz], Phi[fz], Ome[fz] # these are the angles that effectively will be used
        Gc = Gc[fz] # cut away the redundand ones from the rotation array

        # volume element for integrating over the odf
        # see Mason/Patala, arXiv (2019), appendix C
        # omitted factor 1/2 here, since we use omega only from 0 to pi
        dV = np.sin(Ome/2)**2 * np.sin(Tta) * dchi**3 

    # import matplotlib.pyplot as plt
    # fig = plt.figure(figsize=(15, 10))
    # scatter_kwargs = dict(
    #     projection="rodrigues",
    #     figure=fig,
    #     wireframe_kwargs=dict(color="k", linewidth=1, alpha=0.1),
    #     s=5,
    # )
    # ori_plain = Orientation(Rotation([rot.QfromOTP( Gc)]), symmetry=pg).get_random_sample(10000)
    # ori_orix = Orientation(rot_orix, symmetry=pg).get_random_sample(10000)
    # ori_orix.scatter(position=231, c="C0", **scatter_kwargs)
    # ori_plain.scatter(position=232, c="C1", **scatter_kwargs)
    # # ori_quat2.scatter(position=233, c="C2", **scatter_kwargs)

    # ori_orix.scatter(position=234, c="C0", **scatter_kwargs)
    # ori_plain.scatter(position=235, c="C1", **scatter_kwargs)
    # # ori_quat2.scatter(position=236, c="C2", **scatter_kwargs)

    # titles = ["cubochoric", "plain"]#, "quaternion"]
    # for i, title in zip([0, 1], titles):
    #     fig.axes[i].view_init(elev=90, azim=0)
    #     fig.axes[i].set_title(titles[i])
    # for i in [2, 3]:
    #     fig.axes[i].view_init(elev=0, azim=0)

    # volume of the fundamental zone [rad^3]
    V_fz = dV.sum()
    return Gc, dV, V_fz

def qchi(q,chi):
    """ Makes arrays with q and chi values for each detector point

    Parameters
    ----------
    q : 1D ndarray
        momentum exchange
    chi : 1D ndarray
        azimuthal angle on detector

    returns: 
    ------------
    Q, Chi : 1D ndarray, float
        flattened meshgrids of q and chi, representing all detector points
    detshape : tuple, int
        number of q and phi values
    """
    QQ, CHi = np.meshgrid(q,chi)
    detshape = CHi.shape
    Chi, Q = CHi.flatten(), QQ.flatten()
    return Q, Chi, detshape

"""
Some functions helping to convert between different rotation conventions,
to perform rotations and to handle quaternions
"""

def XAfromOTP(ome,tta,phi):
    """ Creates an axis-angle vector from 3 angles omega, theta, phi

    Parameters
    ----------
    omega : ndarray, float
        rotation angle, usually ∈[0,pi)
    theta : ndarray, float
        polar angle of the rotation axis, usually ∈[0,pi)
    phi : ndarray, float
        azimutal angle of the rotation axis, usually ∈[0,2pi)

    Return values
    ------------
    XA : 2D ndarray, float
        array of axis-angle vectors with 3 entries
    """
    XA = np.column_stack([
            ome * np.sin(tta) * np.cos(phi),
            ome * np.sin(tta) * np.sin(phi),
            ome * np.cos(tta)
        ])
    return XA

def OTPfromXA(V):
    """ Extracts the 3 angles omega, theta, phi from an axis-angle vector

    Parameters
    ----------
    XA : 2D ndarray, float
        array of axis-angle vectors with 3 entries

    Return values
    ------------
    omega : ndarray, float
        rotation angle, usually ∈[0,pi)
    theta : ndarray, float
        polar angle of the rotation axis, usually ∈[0,pi)
    phi : ndarray, float
        azimutal angle of the rotation axis, usually ∈[0,2pi)
    """
    ome = np.sqrt((V**2).sum(axis=1))
    tta = np.arccos(np.divide(V[:,2],ome, where=(ome!=0)))
    phi = np.sign(V[:,1]) * np.arccos(
        np.divide(V[:,0],np.sqrt(V[:,0]**2+V[:,1]**2), where=[a[0]!=0 or a[1]!=0 for a in V]) )
    return ome, tta, phi

def RodfromOTP(OTP):
    """ Creates a Rodrigues vector from 3 angles omega, theta, phi

    Parameters
    ----------
    omega : ndarray, float
        rotation angle, usually ∈[0,pi)
    theta : ndarray, float
        polar angle of the rotation axis, usually ∈[0,pi)
    phi : ndarray, float
        azimutal angle of the rotation axis, usually ∈[0,2pi)

    Return values
    ------------
    Rod : 2D ndarray, float
        array of Rodrigues vectors with 3 entries
    """
    ome, tta, phi = OTP[:,0], OTP[:,1], OTP[:,2]
    Rod = np.column_stack([
            np.tan(ome/2) * np.sin(tta) * np.cos(phi),
            np.tan(ome/2) * np.sin(tta) * np.sin(phi),
            np.tan(ome/2) * np.cos(tta)
        ])
    return Rod

@njit#(parallel=True)
def QfromOTP( OTP ):
    """ Creates Quaternions from 3 angles omega, theta, phi

    Parameters
    ----------
    OTP : 2D ndarray, float
        stack of angles for axis-angle rotations
        dim 0: rotation index, 1: [omega, theta, phi]

    Return values
    ------------
    Q : 2D ndarray, float
        stack of quaternions with 4 entries
        dim 0: rotation index, 1: [q0, q1, q2, q3]
    """
    Q = np.empty( (OTP.shape[0],4), data_type )
    for g in prange(OTP.shape[0]):
        ome, tta, phi = OTP[g,0], OTP[g,1], OTP[g,2]
        Q[g] = np.array([
                np.cos(ome/2),
                np.sin(ome/2) * np.sin(tta) * np.cos(phi),
                np.sin(ome/2) * np.sin(tta) * np.sin(phi),
                np.sin(ome/2) * np.cos(tta)
            ])
    Q[Q[:,0] < 0] *= -1
    return Q

@njit#(parallel=True)
def OTPfromQ( Q ):
    """ Extracts the 3 angles omega, theta, phi from a quaterion

    Parameters
    ----------
    Q : 2D ndarray, float
        stack of quaternions with 4 entries
        dim 0: rotation index, 1: [q0, q1, q2, q3]

    Return values
    ------------
    OTP : 2D ndarray, float
        stack of angles for axis-angle rotations
        dim 0: rotation index, 1: [omega, theta, phi]
    """
    OTP = np.empty( (Q.shape[0],3), data_type )
    for g in prange(Q.shape[0]):
        r = np.sqrt( (Q[g,1:]**2).sum() )
        OTP[g] = np.array([
            2 * np.arctan2( r, Q[g,0] ),
            np.arctan2( np.sqrt( Q[g,1]**2 + Q[g,2]**2 ), Q[g,3] ),
            np.arctan2( Q[g,2], Q[g,1] ),
        ])
    OTP[ OTP[:,2]<0, 2 ] += 2*np.pi
    return OTP
        
@njit
def MatrixfromOTP( ome, tta, phi ):
    """ Creates rotation matrices from 3 angles omega, theta, phi

    Parameters
    ----------
    omega : ndarray, float
        rotation angle, usually ∈[0,pi)
    theta : ndarray, float
        polar angle of the rotation axis, usually ∈[0,pi)
    phi : ndarray, float
        azimutal angle of the rotation axis, usually ∈[0,2pi)

    Return values
    ------------
    R : ndarray, float
    """
    x = np.sin(tta) * np.cos(phi)
    y = np.sin(tta) * np.sin(phi)
    z =  np.cos(tta)
    so = np.sin(ome)
    co = np.cos(ome)
    R = np.array([
        [ co + x**2*(1-co), x*y*(1-co) - z*so, x*z*(1-co) + y*so ],
        [ y*x*(1-co) + z*so, co + y**2 *(1-co), y*z*(1-co) - x*so ],
        [ z*x*(1-co) - y*so, z*y*(1-co) + x*so, co + z**2 *(1-co) ]
    ])
    return R.astype(data_type)

def EulerfromQ(Q):
    """
    Convert a quaternion to Bunge Euler angles (phi1, Phi, phi2).
    Quaternion q = (w, x, y, z) must be normalized.
    Returns angles in radians.
    """
    Euler = np.empty_like(Q[:,:3])
    for k, q in enumerate(Q):
        w, x, y, z = q

        # Normalize to avoid numerical drift
        norm = np.sqrt(w*w + x*x + y*y + z*z)
        w, x, y, z = w/norm, x/norm, y/norm, z/norm

        # Rotation matrix from quaternion
        R = np.array([
            [1 - 2*(y*y + z*z), 2*(x*y - z*w), 2*(x*z + y*w)],
            [2*(x*y + z*w), 1 - 2*(x*x + z*z), 2*(y*z - x*w)],
            [2*(x*z - y*w), 2*(y*z + x*w), 1 - 2*(x*x + y*y)]
            ])

        # Bunge convention (ZXZ intrinsic rotations)
        Phi = np.arccos(np.clip(R[2,2], -1.0, 1.0))

        if abs(Phi) < 1e-8:
            phi1 = np.arctan2(-R[0,1], R[0,0])
            phi2 = 0.0
        elif abs(Phi - np.pi) < 1e-8:
            phi1 = np.arctan2(R[0,1], -R[0,0])
            phi2 = 0.0
        else:
            phi1 = np.arctan2(R[2,0], -R[2,1])
            phi2 = np.arctan2(R[0,2], R[1,2])

        # Wrap to [0, 2π)
        Euler[k,0] = phi1 % (2*np.pi)
        Euler[k,1] = Phi % np.pi
        Euler[k,2] = phi2 % (2*np.pi)

    return Euler

@njit
def stack_mrot( vs, OTP ):
    """ Rotates a stack of vectors by a list of rotations

    Parameters
    ----------
    vs : 2D ndarray, float
        stack of vectors
        dim 0: vector index, 1: [x,y,z]
    OTP : 2D ndarray, float
        stack of angles for axis-angle rotations
        dim 0: rotation index, 1: [omega, theta, phi]

    Return values
    ------------
    R : ndarray, float
    """
    Nrot = OTP.shape[0]
    vs_rot = np.empty( ( Nrot, vs.shape[0], 3 ), data_type)
    for g in prange( Nrot ):
        R = MatrixfromOTP( OTP[g,0], OTP[g,1], OTP[g,2] ).T
        vs_rot[g] = vs @ R
    return vs_rot

@njit(parallel=True)
def ang_distance( q1, q2, gen ):
    """ Calculates the closest distance between 2 quaternions
    respecting a crystal symmetry represented by symmetry generators

    Q. Huynh, Metrics for 3d rotations: Comparison and analysis, 
    Journal of Mathematical Imaging and Vision, vol. 35, pp. 155-164, 2009.

    Parameters
    ----------
    q1, q2 : 2D ndarray, float
        stacks of quaternions
    gen : 2D ndarray, float
        2 symmetry generators in OTP representation
        [[ome1,tta1,phi1], [ome2,tta2,phi2]]

    Return values
    ------------
    DQ : ndarray, float
    """
    DQ = np.empty(q1.shape[0], data_type )
    # make quaternions from generators
    Q_gen = QfromOTP( gen )
    n1 = int(2*np.pi / gen[0,0]) # n1-fold symmetry from 1st generator
    n2 = int(2*np.pi / gen[1,0]) # n2-fold symmetry from 2nd generator
    for k in prange(q1.shape[0]):
        dq = np.pi
        qc = q2[k]
        for _ in range(n1):
            qc = quaternion_multiply( qc, Q_gen[0] ) # rotate point by 1st generator
            for _ in range(n2):
                qc = quaternion_multiply( qc, Q_gen[1] ) # rotate point by 2st generator
                # add to the odf according to the real quaternion distance
                r = 2*(qc*q1[k]).sum()**2 - 1
                if np.abs(r) > 1:
                    r /= np.abs(r) # this is for rounding errors
                dq = min(dq, np.arccos( r ))
        DQ[k] = dq
    return DQ

@njit
def symmetry_equivalent_quaternions(q, Q_group, prec=7):
    Q_eq = np.empty_like(Q_group)
    for k in range(Q_group.shape[0]):
        Q_eq[k] = np.round( quaternion_multiply( Q_group[k], q ), prec)

    Q_eq[Q_eq[:,0]<0] *= -1 # bring all to positive real part
    return nb.nb_unique_axis0(Q_eq)

@njit(parallel=True)
def misorientation( q1, q2, gen ):
    """ Calculates the closest distance between 2 quaternions
    respecting a crystal symmetry represented by symmetry generators

    Q. Huynh, Metrics for 3d rotations: Comparison and analysis, 
    Journal of Mathematical Imaging and Vision, vol. 35, pp. 155-164, 2009.

    Parameters
    ----------
    q1, q2 : 2D ndarray, float
        stacks of quaternions
    gen : 2D ndarray, float
        2 symmetry generators in OTP representation
        [[ome1,tta1,phi1], [ome2,tta2,phi2]]

    Return values
    ------------
    DQ : ndarray, float
    """
    DQ = np.empty(q1.shape[0], data_type )
    # make quaternions from generators
    Q_gen = QfromOTP( gen )
    if gen[0,0] > 0:
        n1 = int(2*np.pi / gen[0,0]) # n1-fold symmetry from 1st generator
    else:
        n1=1
    if gen[1,0] > 0:
        n2 = int(2*np.pi / gen[1,0]) # n2-fold symmetry from 2nd generator
    else:
        n2=1
    for k in prange(q1.shape[0]):
        dq = np.pi
        qa = q1[k]
        qb = q2[k]
        for _ in range(n1):
            qa = quaternion_multiply( Q_gen[0], qa ) # rotate point by 1st generator
            for _ in range(n2):
                qa = quaternion_multiply( Q_gen[1], qa ) # rotate point by 2st generator
                for _ in range(n1):
                    qb = quaternion_multiply( qb, Q_gen[0] ) # rotate point by 1st generator
                    for _ in range(n2):
                        qb = quaternion_multiply( qb, Q_gen[1] ) # rotate point by 2st generator
                        # add to the odf according to the real quaternion distance
                        r = 2*(qa*qb).sum()**2 - 1
                        if np.abs(r) > 1:
                            r /= np.abs(r) # this is for rounding errors
                        dq = min(dq, np.arccos( r ))
        DQ[k] = dq
    return DQ

@njit()
def misorientation_single( qa, qb, gen ):
    """ Calculates the closest distance between 2 quaternions
    respecting a crystal symmetry represented by symmetry generators

    Q. Huynh, Metrics for 3d rotations: Comparison and analysis, 
    Journal of Mathematical Imaging and Vision, vol. 35, pp. 155-164, 2009.

    Parameters
    ----------
    q1, q2 : 1D ndarray, float
        quaternions to compare
    gen : 2D ndarray, float
        2 symmetry generators in OTP representation
        [[ome1,tta1,phi1], [ome2,tta2,phi2]]

    Return values
    ------------
    DQ : ndarray, float
    """
    # make quaternions from generators
    Q_gen = QfromOTP( gen )
    n1 = int(2*np.pi / gen[0,0]) # n1-fold symmetry from 1st generator
    n2 = int(2*np.pi / gen[1,0]) # n2-fold symmetry from 2nd generator
    dq = np.pi
    for _ in range(n1):
        qa = quaternion_multiply( Q_gen[0], qa ) # rotate point by 1st generator
        for _ in range(n2):
            qa = quaternion_multiply( Q_gen[1], qa ) # rotate point by 2st generator
            for _ in range(n1):
                qb = quaternion_multiply( qb, Q_gen[0] ) # rotate point by 1st generator
                for _ in range(n2):
                    qb = quaternion_multiply( qb, Q_gen[1] ) # rotate point by 2st generator
                    # add to the odf according to the real quaternion distance
                    r = 2*(qa*qb).sum()**2 - 1
                    if np.abs(r) > 1:
                        r /= np.abs(r) # this is for rounding errors
                    dq = min(dq, np.arccos( r ))
    return dq

@njit
def quaternion_multiply(q1, q2):
    """ Multiplies 2 quaternions according to their rules

    Parameters
    ----------
    q1, q2 : ndarray, float
        quaternions [q0, q1, q2, q3]

    Return values
    ------------
    multiplied quaternion [q0, q1, q2, q3]
    """
    w1, x1, y1, z1 = q1
    w2, x2, y2, z2 = q2
    w = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2
    x = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2
    y = w1 * y2 + y1 * w2 + z1 * x2 - x1 * z2
    z = w1 * z2 + z1 * w2 + x1 * y2 - y1 * x2
    return np.array([w, x, y, z], data_type)

@njit()
def quaternion_rotate_vector(q, v):
    """Rotate vector v by quaternion q."""
    q_conj = np.array([q[0], -q[1], -q[2], -q[3]])
    v_quat = np.array([0.0, *v])
    rotated = quaternion_multiply(
        quaternion_multiply(q, v_quat),
        q_conj
    )
    return rotated[1:]  # return x, y, z

@njit()
def quaternion_inverse_rotate_vector(q, v):
    """Rotate vector v by quaternion q."""
    q_conj = np.array([q[0], -q[1], -q[2], -q[3]])
    v_quat = np.array([0.0, *v])
    rotated = quaternion_multiply(
        quaternion_multiply(q_conj, v_quat),
        q
    )
    return rotated[1:]  # return x, y, z

def mean_orientation(Qc, odf):
    """Quaternion Averaging via Eigen Decomposition (Markley et al. 2007)
    Works only in full SO(3), not respecting symmetries.

    Parameters
    ----------
    Qc : 2d ndarray (N,4)
        array of quaternions 
    odf : 1d ndarray (N)
        weight for each quaternion

    Returns
    -------
    ndarray (4)
        average quaternion
    """
    # Handle sign ambiguity
    for i in range(1, Qc.shape[0]):
        if np.dot(Qc[0], Qc[i]) < 0:
            Qc[i] = -Qc[i]

    # Build the symmetric accumulator matrix
    A = np.zeros((4,4),data_type)
    for k, o in enumerate(odf):
        A += o * np.outer(Qc[k],Qc[k])
    A /= odf.sum()
    # Compute the eigenvector corresponding to the largest eigenvalue
    eigvals, eigvecs = np.linalg.eigh(A)
    avg_quaternion = eigvecs[:, np.argmax(eigvals)]

    return -avg_quaternion

def samplerotations_eq3D( d_rot, tilts ):
    """ Calculates rotation and tilt angles equally spaced over a sphere

    Parameters
    ----------
    d_rot : float
        distance between rotations at zero tilt in degree
    tilts : list of int
        list of tilt angles in degree

    returns: 
    ------------
    Omega, Kappa : ndarray, float
        rotation and tilt angles in degree
    """
    omega, kappa = np.array([]),np.array([])
    d_ome = d_rot
    for k, kap in enumerate( np.atleast_1d(tilts) ):
        if kap == 0.0:
            ome_max = 180
        else:
            ome_max = 360
        n_ome = np.round(ome_max/d_ome * np.cos(kap*np.pi/180)).astype(int)
        if np.mod(k,2) != 0:
            ome_min = ome_max/n_ome/2
        else:
            ome_min = 0
        omega = np.concatenate(
            [omega, np.linspace(ome_min,ome_max,n_ome, endpoint=False)],
            axis=0 )
        kappa = np.concatenate([kappa, kap*np.ones(n_ome)], axis=0)

    return np.array(omega).flatten(), np.array(kappa).flatten()

def generate_group(generators, prec = 7):
    G = set()
    queue = [tuple(g) for g in generators]
    G.add((1.0, 0.0, 0.0, 0.0))  # Identity

    while queue:
        current = queue.pop()
        for g in G.copy():
            new = quaternion_multiply(np.array(current), np.array(g))
            if new[0] < 0:
                new *= -1
            new = new / np.linalg.norm(new)
            tup = tuple(np.round(new, decimals=prec))  # avoid FP errors
            if tup not in G:
                G.add(tup)
                queue.append(tup)
            if len(queue) > 40:
                prec -= 1
                queue = round_remove_duplicates(queue, prec)

    return np.array(list(G), data_type)

def round_remove_duplicates(array_list, decimals=3):
    rounded = [np.round(arr, decimals) for arr in array_list]
    
    seen = set()
    unique = []
    for arr in rounded:
        key = tuple(arr.flatten())
        if key not in seen:
            seen.add(key)
            unique.append(arr)
    
    return unique

@njit
def symmetry_equivalent_quaternions(q, Q_group, prec=7):
    Q_eq = np.empty_like(Q_group)
    for k in range(Q_group.shape[0]):
        Q_eq[k] = np.round( quaternion_multiply( Q_group[k], q ), prec)

    Q_eq[Q_eq[:,0]<0] *= -1 # bring all to positive real part
    return nb.nb_unique_axis0(Q_eq)

def map_to_fz(q, Q_group, q_ref=np.array([1, 0, 0, 0])):
    q = q / np.linalg.norm(q)
    best_q = q
    max_dot = -1
    for q_op in Q_group:
        q_sym = quaternion_multiply(q_op, q)
        dot = abs(np.dot(q_sym, q_ref))
        if dot > max_dot:
            max_dot = dot
            best_q = q_sym
    if best_q[0] < 0:
        best_q *= -1
    return best_q / np.linalg.norm(best_q)

@njit()
def quaternion_conjugate(q):
    return np.array([q[0], -q[1], -q[2], -q[3]])

@njit()
def misorientation_angle(q1, q2, Q_group):
    """ Calculates the closest distance between 2 quaternions symmetry
    given by a point group
    
    newer version using the full proper point group """
    q1 = q1 / np.linalg.norm(q1)
    q2 = q2 / np.linalg.norm(q2)
    w_max = 0.
    for q_op in Q_group:
        q2_rot = quaternion_multiply(q_op, q2)
        q_rel = quaternion_multiply(quaternion_conjugate(q1), q2_rot)
        w = nb.nb_clip(abs(q_rel[0]), 0.0, 1.0)
        w_max = max(w,w_max)
    return 2 * np.arccos(w_max)  # in radians

@njit()#parallel=True)
def misorientation_angle_stack(Q1, Q2, Q_group):
    mori_stack = np.empty_like(Q1[:,0])
    for k in prange(Q1.shape[0]):
        mori_stack[k] = misorientation_angle(Q1[k],Q2[k],Q_group)
    return mori_stack

@njit()
def slerp(q0, q1, t, eps=1e-6):
    # spline linear interpolation between 2 quaternions for t in [0,1]
    d = np.dot(q0,q1)
    if d < 0:
        q1 = -q1
        d = -d
    if d > 1 - eps:
        # nearly identical -> nlerp
        q = (1-t)*q0 + t*q1
        return q / quaternion_norm(q)
    Omega = np.arccos(d)
    s0 = np.sin((1-t)*Omega) / np.sin(Omega)
    s1 = np.sin(t*Omega) / np.sin(Omega)
    return s0*q0 + s1*q1

@njit()
def nlerp(q0, q1, t):
    # normalized linear interpolation (faster) between 2 quaternions for t in [0,1]
    q = (1-t)*q0 + t*q1
    return q / quaternion_norm(q)

@njit
def quaternion_norm(q):
    """Gives the vectornorm along the first axis
    """
    norm = np.sqrt(q[0]*q[0] + q[1]*q[1] + q[2]*q[2] + q[3]*q[3])
    return norm

@njit
def quaternion_dot(q1,q2):
    """Gives the dot product along the first axis
    """
    return q1[0]*q2[0] + q1[1]*q2[1] + q1[2]*q2[2] + q1[3]*q2[3]


import os, glob
import numpy as np
from time import time
import sys
import h5py
from numba import prange, njit
from scipy.signal import find_peaks

# domestic
from .. import mask as msk
from . import baselines as bln
from .. import numba_plugins as nb
from ..misc import import_module_from_path, find_unique_positions
from .. import exceptions as exc
from ..model.model_textom import model_textom
from ...config import data_type
from ...version import __version__

def import_data( sample_dir, pattern, mod:model_textom,  baseline_path,
                peak_region_path, blmask_path, detmask_path, geo_path,
                use_ion=True ):
    """Looks for data in path/data_integrated/ and prepares them for textom reconstructions

    Parameters
    ----------
    sample_dir : str
        textom base directory
    pattern : str
        substring required to be in the files in path/data_integrated/
    mod : model
        model object, needs to have projectors and diffractlets
    baseline_path : str 
        path to the desired background_subtraction module
    qmask_path : str, optional
        path to a file containing the peak-regions in q, if None will be created
        from user input, by default None
    detmask_path : str, optional
        path to a file containing the detector mask, if None will be created
        from user input, by default None
    geo_path : str, optional
        path to the desired geometry module
    flip_fov : bool, optional
        can be set to True if the fov metadata is switched by accident, by default False
    use_ion : bool, optional
        choose if normalisation by ionization chamber should be used (if present in
        data), by default True
    """
    # Load stuff
    geo = import_module_from_path('geometry', geo_path)
    bl = import_module_from_path('background_subtraction', baseline_path)
    peak_reg, hkl = load_peak_regions(peak_region_path, exclude_empty=True)
    bl_reg, _ = load_peak_regions(blmask_path, exclude_empty=False)
    mask_cake = h5py.File(detmask_path,'r')['mask_cake'][()]
    '''
    # sanity check compare number of q-values in mask and peak-regions:
    n_chi = mod.detShape[0]
    n_q_detmask = mask_detector.size//n_chi
    # if not n_q_detmask == peak_reg.shape[0]:
    #     raise exc.dimension_error('Number of q-values in peak regions and detector mask do not coincide.')
    # # reshape mask_detector because q and chi axes are defined the other way during masking
    # mask_detector = mask_detector.reshape((n_q_detmask, n_chi)).swapaxes(0,1).flatten()
    # i'll not change the masking now because it would be better to import the whole data
    # and write them in 2d to data_textom
    # and then do the masking afterwards in fitting.py
    # anyway this also replaces the sanity check above
    '''

    # get the images that show the sample
    scanmask = mod.Beams[:,:,0].astype(bool)                           ###########

    print('Starting data import')
    print('\tLoading integrated data from files')
    t0=time()
    airscat, ion = [], []
    try:
        filelist = h5py.File(os.path.join(sample_dir,'analysis','data_mumott.h5'), 'r')['file_names'][()]
    except:
        sort_by_angles = False if geo.scan_mode == 'controt' else True
        filelist = get_data_list(sample_dir, pattern, sort_by_angles=sort_by_angles)

    if 'controt' in geo.scan_mode:
        # continuos rotation import
        all_data = []
        y_positions, z_positions = [], []
        for f, file in enumerate(filelist):
            with h5py.File(file, 'r') as hf:
                if 'ion' in hf and use_ion:
                    try:
                        ion_g = hf['ion'][()]
                        ion_g[ion_g>1] = winsorize(ion_g[ion_g>1], 1)
                    except:
                        ion_g = hf['ion'][()]
                    ion.append(ion_g)
                else:
                    use_ion=False
                if f==0:
                    q_in = np.squeeze(hf['radial_units'][()] ).astype(data_type)
                    chi_in = np.squeeze( hf['azimuthal_units'][()]*np.pi/180 ).astype(data_type)
                    omega = ( np.array(hf['rot_angle'][()]) ).astype(data_type)
                    kappa = np.zeros_like(omega)

                y_positions.append( np.mean(hf['ty'][()]) )
                z_positions.append( hf['tz'][()] )    
                d = hf['cake_integ'][()]

            # rescale data by primary beam intensity if known
            if use_ion:
                if g == 0:
                    scale = np.mean(ion[0])
                d = _rescale( d, ion[g] ) * scale
            all_data.append( d )

        # # find out y and z grids                   
        y_indices = find_unique_positions(y_positions)
        z_indices = find_unique_positions(z_positions)

        n_rot_angles = omega.size
        n_ybins = y_indices.max() + 1
        n_zbins = z_indices.max() + 1
        fov = (n_ybins, n_zbins)
        projections = np.zeros((n_rot_angles, n_ybins, n_zbins, *all_data[0].shape[1:]), data_type) # check size and see if this would be good for preprocess_data()
        for k,d in enumerate(all_data):
            projections[:,y_indices[k],z_indices[k],...] = d

        projections = projections.reshape((n_rot_angles, n_ybins*n_zbins, *all_data[0].shape[1:]))
    # # continuos rotation import
    # all_data = []
    # y_positions, z_positions = [], []
    # for f, file in enumerate(filelist):
    #     with h5py.File(file, 'r') as hf:
    #         if f==0:
    #             q_in = np.squeeze(hf['radial_units'][()] ).astype(data_type)
    #             chi_in = np.squeeze( hf['azimuthal_units'][()]*np.pi/180 ).astype(data_type)
    #             rot_angles = ( np.array(hf['rot_angle'][()]) ).astype(data_type)
    #         y_positions.append( hf['ty'][()] )
    #         z_positions.append( hf['tz'][()] )    
    #         all_data.append( ( np.array(hf['cake_integ'][()]) ).astype(data_type) )

    cp_threshold = False # for crazy pixel filter
    for g, file in enumerate( filelist ):
        if 'controt' in geo.scan_mode:
            proj = projections[g]
        else:
            # Read data from integrated file
            with h5py.File(os.path.join(sample_dir, 'data_integrated', file),'r') as hf:
                q_in = np.squeeze(hf['radial_units'][()] ).astype(data_type)
                chi_in = np.squeeze( hf['azimuthal_units'][()]*np.pi/180 ).astype(data_type)
                fov = ( hf['fov'][()] ).astype(np.int32)
                d = ( np.array(hf['cake_integ'][()]) ).astype(data_type)
                if 'ion' in hf and use_ion:
                    try:
                        ion_g = hf['ion'][()]
                        ion_g[ion_g>1] = winsorize(ion_g[ion_g>1], 1)
                    except:
                        ion_g = hf['ion'][()]
                    ion.append(ion_g)
                else:
                    use_ion=False

                ty = np.atleast_1d(hf['ty'][()])
                y_scan_direction = np.sign(ty[-1] - ty[0])
                tz = np.atleast_1d(hf['tz'][()])
                z_scan_direction = np.sign(tz[-1] - tz[0])

            # this is just for the case fast/slow axes were chosen wrong during integration
            if geo.flip_fov:
                fov = np.flip( fov )
                
            # rescale data by primary beam intensity if known
            if use_ion:
                if g == 0:
                    scale = np.mean(ion[0])
                d = _rescale( d, ion[g] ) * scale

            ## Reshaping in function of scanning mode ###
            # Base code written for column scan
            d = d.reshape( *fov, d.shape[1], d.shape[2] )
            
            if 'line' in geo.scan_mode:
                fov = np.flip( fov )
                # d = np.fliplr(np.flipud( np.transpose( d, axes=(1,0,2,3)) ) )
                d = np.transpose( d, axes=(1,0,2,3)) 

            # reorder data so that they are all treated the same in the model
            if 'snake' in geo.scan_mode:
                # Flip every second row
                for ii in range(d.shape[0]):
                    if ii % 2 != 0:
                        d[ii] = d[ii][::-1]

            if y_scan_direction < 0:
                d = np.flip(d, axis=0)
            if z_scan_direction < 0:
                d = np.flip(d, axis=1)

            ##############################################

            i0=d.shape[-1]//4
            half_fov = fov[1]//2
            edge_sample = np.array([ 
                d[0,0,i0:].mean(),d[0,-1,:,i0:].mean(), 
                d[0,-half_fov,:,i0:].mean(),  d[-1,0,:,i0:].mean(), 
                d[-1,-half_fov,:,i0:].mean(),d[-1,-1,:,i0:].mean()]) 
            airscat.append( max( np.min( edge_sample), 0.) )
            # rescale by airscattering if primary beam intensity not known
            if not ion or not use_ion:
                # print('\tRescale by air scattering')
                if g==0:
                    scale = airscat[0]
                else:
                    d *= (scale / airscat[g])

            # pad the data as in mumottize
            n_chi = chi_in.size
            proj =  np.zeros( (*mod.fov, n_chi, *q_in.shape), data_type)
            si, sj = d.shape[:2]
            i0 = (mod.fov[0] - si)//2
            j0 = (mod.fov[1] - sj)//2
            proj[i0:i0+si, j0:j0+sj] = d
            proj = proj.reshape(mod.fov[0]*mod.fov[1], n_chi, *q_in.shape)
        
        if not cp_threshold: # like this it's done only once
            print('\tChoose threshold for crazy pixels (Everything below will be processed)')
            cp_threshold = msk.select_threshold_hist( proj.flatten(), 
                        xlabel='Counts', ylabel='No of data points', 
                        title='Choose threshold for crazy pixels (Everything below will be processed)',
                        logx=True, logy=True )
            
            if bl.mode == 'custom':
                custom_bl = True
                with h5py.File(os.path.join(sample_dir,bl.path_custom_bl), 'r') as hf:
                    baseline_values = hf[bl.h5path_custom_bl][()]
                _, q_mask_p = get_q_mask(q_in, peak_reg)
                bl_args = [bl.path_custom_bl, bl.h5path_custom_bl]
            else:
                custom_bl = False
                _, q_mask_p = get_q_mask(q_in, peak_reg)
                q_mask_bl, _ = get_q_mask(q_in, bl_reg)
                match bl.mode:
                    case 'linear'|'linear_azimuthal':
                        print('\tSubtracting linear baselines')
                        bl_fun = bln.linear_baseline
                        bl_args = (0)
                    case 'chebyshev'|'chebyshev_azimuthal':
                        print('\tSubtracting chebyshev polynomial baselines')
                        bl_fun = bln.chebyshev_baseline
                        bl_args = (bl.order_chebyshev)
                    # case 'chebyshev_auto':
                    #     print('\tSubtracting chebyshev polynomial baselines with automated masking')
                    #     bl_fun = bln.auto_chebyshev
                    #     bl_args = (bl.order_chebyshev, bl.pre_order, bl.k_sigma, bl.q_expand)
                    case 'none':
                        print('\tNot subtracting baselines')
                        bl_fun = bln.no_baseline
                        bl_args = (0)
                    case _:
                        print('\tBaseline mode not recognized, revise background_subtraction.py')
                        break
                t1 = time()

        # remove hot pixels / all values above threshold
        proj[proj > cp_threshold] = np.median(proj)

        # subtract baselines and regroup data into the chosen q-bins
        t_mask = np.where(scanmask[g])[0]
        if custom_bl:
            data_fit =  _regroup_q_custom_baseline( 
                proj, t_mask, mask_cake, baseline_values[g], q_mask_p )
        elif 'azimuthal' in bl.mode:
            data_fit = _regroup_q_azimuthally_resolved_baseline( 
                proj, t_mask, mask_cake,
                bl_fun, bl_args, q_in, q_mask_bl, q_mask_p )
        else:
            data_fit = _regroup_q_baseline( 
                proj, t_mask, mask_cake,
                bl_fun, bl_args, q_in, q_mask_bl, q_mask_p )
        
        out_path = os.path.join( sample_dir, 'analysis', 'data_textom.h5')
        if g == 0:
            # prepare mask on chi and q
            QQ,_ = np.meshgrid( q_in, chi_in )
            qmask = [np.logical_and(QQ >= start, QQ <= end) for start,end in peak_reg]
            mask_peaks = np.array([
                [np.logical_or.reduce(mask_cake[k, qmask[l][k]]) for l in range(len(peak_reg))] 
                    for k in range(chi_in.size)])

            print('\tSaving data to file: %s' % out_path)
            with h5py.File( out_path, 'w') as hf:
                hf.create_dataset( 'data',
                        shape=(0, data_fit.shape[1], data_fit.shape[2]),
                        maxshape=(None, data_fit.shape[1], data_fit.shape[2]),
                        chunks=(1, data_fit.shape[1], data_fit.shape[2]),
                        dtype=data_type,
                    )
                hf.create_dataset( 'peak_reg', data=peak_reg )
                hf.create_dataset( 'q', data = np.mean(peak_reg, axis=1))
                hf.create_dataset( 'chi', data = chi_in )
                hf.create_dataset( 'hkl', data=hkl )
                hf.create_dataset( 'detShape', data = [n_chi, q_mask_p.shape[0]] )
                hf.create_dataset( 'mask_cake', data = mask_cake)
                hf.create_dataset( 'q_in', data = q_in )
                hf.create_dataset( 'mask_peaks', data = mask_peaks )
                hf.create_dataset( 'file_names', data = filelist )
                
        # add data to h5 file
        with h5py.File( out_path, 'r+') as hf:
            dset = hf['data']
            current_rows = dset.shape[0]
            new_rows = current_rows + data_fit.shape[0]
            dset.resize( new_rows, axis=0)
            dset[current_rows:new_rows] = data_fit

        try:
            t_it = (time()-t1)/(g)
        except:
            t_it=0
        Nrot = len(filelist)
        sys.stdout.write(f'\r\t\tProjection {(g+1):d} / {Nrot:d}, t/proj: {t_it:.1f} s, t left: {((Nrot-g-1)*t_it/60):.1f} min' )
        sys.stdout.flush()

    gt_mask = np.array( np.where(scanmask) ).T
    # add metadata to h5 file
    with h5py.File( out_path, 'r+') as hf:
        hf.create_dataset( 'scanmask', data=scanmask )
        hf.create_dataset( 'gt_mask', data=gt_mask )
        hf.create_dataset( 'baseline_mode', data=bl.mode )
        hf.create_dataset( 'baseline_order', data=bl_args )        
        hf.create_dataset( 'airscat', data = airscat )
        if ion:
            ion_av = np.array( [np.mean(i) for i in ion] ) # cannon write ion directly because shape
            hf.create_dataset( 'ion', data = ion_av )

    print(f'\n\t\ttook {(time()-t0)/60:.1f} min')

def get_data_list(sample_dir, pattern, subdir='data_integrated', sort_by_angles=True):
    """Looks for all data in data_integrated and makes a list. 
    One can choose a pattern include only the desired files.

    Parameters
    ----------
    sample_dir : str
        sample base directory
    pattern : str
        integrated data filename with * placeholders

    Returns
    -------
    list
        list of filenames to process
    """
    filelist = sorted( glob.glob(os.path.join(sample_dir,subdir,'*h5')) )
    # the exception is for the coefficients for generated samples
    filelist = [s for s in filelist if pattern in s]#[:2]
    filelist = [s for s in filelist if 'sample_coeff' not in s]
    if sort_by_angles:
        try:
            filelist = sort_data_list_by_angles( filelist, 'tilt_angle', 'rot_angle' )
            return filelist
        except:
            raise exc.metadata_missing('Did not find rotation and tilt angles in integrated data files')
    else:
        return filelist 

def mask_peak_regions( mod:model_textom, q_data, powder_1D, peak_reg_path):
    # t_inter = time() # this is to subtract the time it takes for user input for remaining time estimation
    # set up boolean mask for filtering data
    q = mod.Qq_det.reshape(mod.detShape)[0]
    q_mask = np.ones_like(q_data, dtype=bool)
    # select peak regions from data and simulated powder pattern
    print('\tChoose regions containing Bragg peaks, then close the figure')
    # happy = 'n'
    # while happy != 'y':
    powder = mod.powder_pattern * powder_1D.max()/mod.powder_pattern.max()

    peak_reg = msk.select_regions( q_data[q_mask], powder_1D[q_mask], q, powder, hkl=mod.hkl,
            max_regions=None,
            title='Select individual Bragg peaks by holding LMB, remove by RMB' )
        # happy = input('\thappy? (y/n) ')
    peak_reg = peak_reg.get_regions()

    with open( peak_reg_path,'w') as fid:
        for reg in peak_reg:
            peak_hkl = mod.hkl[ np.logical_and( q>=reg[0], q<=reg[1] ) ]
            peak_hkl_str = ",".join("[" + ",".join(map(str, row)) + "]" for row in peak_hkl)
            fid.write(f'{reg[0]}\t{reg[1]}\t{peak_hkl_str}\n')
    print(f'\tSaved peak regions to {peak_reg_path}')

def mask_for_baseline( mod:model_textom, q_data, powder_1D, bl_reg_path):
    # t_inter = time() # this is to subtract the time it takes for user input for remaining time estimation
    # set up boolean mask for filtering data
    q = mod.Qq_det.reshape(mod.detShape)[0]
    q_mask = np.ones_like(q_data, dtype=bool)
    # select peak regions from data and simulated powder pattern
    print('\tSelect regions excluded in baseline subtraction, then close the figure')
    # happy = 'n'
    # while happy != 'y':
    powder = mod.powder_pattern * powder_1D.max()/mod.powder_pattern.max()

    peak_reg = msk.select_regions( q_data[q_mask], powder_1D[q_mask], q, powder, hkl=mod.hkl,
            max_regions=None,
            title='Select regions excluded in baseline subtraction by holding LMB, remove by RMB' )
        # happy = input('\thappy? (y/n) ')
    peak_reg = peak_reg.get_regions()

    with open( bl_reg_path,'w') as fid:
        for reg in peak_reg:
            fid.write(f'{reg[0]}\t{reg[1]}\n')
    print(f'\tSaved regions to {bl_reg_path}')

    # q_peaks = np.mean(peak_reg, axis=1)

    # # find out prominence of the highest peak for filtering data
    # I_mx = 0.                   
    # q_mask_k = []
    # for k, (start, end) in enumerate( peak_reg ):
    #     q_mask_k.append( ((q_data >= start) & (q_data <= end)) )
    #     q_mask &= ~q_mask_k[k]
    #     dat_peak = powder_1D[q_mask_k[k]]
    #     if dat_peak.max() > I_mx:
    #         q_mask_hp = q_mask_k[k]
    #         _, info = find_peaks(dat_peak, prominence=0.)
    #         try:
    #             prom = info['prominences'].max()
    #         except:
    #             prom=0.
    #         I_mx = dat_peak.max()
    # q_mask_k = np.array(q_mask_k)#.astype(data_type)  

    # t_inter = time()-t_inter
    # return peak_reg, q_mask, q_mask_k, q_mask_hp, prom, q_peaks, t_inter

def load_peak_regions( peak_reg_path, exclude_empty=False ):
    peak_reg,hkl = [],[]
    with open(peak_reg_path) as f:
        for line in f:
            # parts = line.strip().split("\t")
            # parts = list(filter(None, parts)) # this removes empty tabs
            parts = [s.strip() for s in line.split()]
            if len(parts) != 3:
                if exclude_empty:
                    continue
                else:
                    parts.append('')
            peak_reg.append((float(parts[0]), float(parts[1])))
            hkl.append(parts[2])

    # peak_reg = np.genfromtxt(peak_reg_path)
    # peak_reg = peak_reg.reshape( (peak_reg.size//2, 2)  )
    return np.array(peak_reg), hkl

def get_powder1d( path_projection, path_cakemask ):
    with h5py.File(path_projection, 'r') as hf:
        data_2d = hf['cake_integ'][()].sum(axis=0) # load data and sum over all good images in the projection
        q_powder = hf['radial_units'][()].flatten()
    if os.path.isfile(path_cakemask): # get the mask from pyfai (might not be completely accurate though)
        with h5py.File(path_cakemask,'r') as hf:
            mc = hf['mask_cake'][()].astype(bool)
    else:
        mc = np.ones_like(data_2d)
    data_2d[~mc] = np.nan # exclude masked values
    powder = np.nansum(data_2d, axis=0) # sum over all angles to get the powder pattern
    return q_powder, powder

def get_q_mask( q, peak_regions ):
    q_mask_p = []
    for (start, end) in peak_regions:
        q_mask_p.append( ((q >= start) & (q <= end)) )
    q_mask_p = np.array(q_mask_p)
    q_mask = np.logical_not( q_mask_p.sum(axis=0) )
    return q_mask, q_mask_p

def mask_detector( cakemask, detmask_path, powder_2D_masked, peak_reg, q_in, chi_in ):

    QQ,_ = np.meshgrid( q_in, chi_in )
    qmask = [np.logical_and(QQ >= start, QQ <= end) for start,end in peak_reg]
    mask_detector = np.array([
        [np.logical_or.reduce(cakemask[k, qmask[l][k]]) for k in range(chi_in.size)] 
            for l in range(len(peak_reg))])

    print('\tCreate mask by removing (left mouse button) or restoring pixels (right mouse button)')
    # start = np.argmax(q_mask) # this could be useful if taking a full mask, then partially regroup ? have to do it with diffractlets too maybe
    # end = len(q_mask) - 1 - np.argmax(q_mask[::-1]) # then make these values None and when regrouping do np.mean * number of points to make up for Nones
    mask_detector = msk.mask_detector( powder_2D_masked, mask_detector )
    with open( detmask_path, 'w' ) as fid:
        for pxl in mask_detector:
            fid.write(f'{pxl}\n')
    return mask_detector

def sort_data_list_by_angles( filelist, h5_outer_rotaxis_path, h5_inner_rotaxis_path ):

    inner_angle = []
    outer_angle = []
    for file in filelist:
        with h5py.File( file, 'r') as hf:
            inner_angle.append( hf[h5_inner_rotaxis_path][()] )
            outer_angle.append( hf[h5_outer_rotaxis_path][()] )
    # first sort by outer angles:
    order = np.lexsort([inner_angle, outer_angle])

    return [filelist[k] for k in order]

def winsorize(data, percent = 1):
    """Very simple filter, setting the upper and lower percentile to the next smaller/larger value
    """
    lower = np.percentile(data, percent)
    upper = np.percentile(data, 100-percent)
    data[data < lower] = lower
    data[data > upper] = upper
    return data

################## numba compiled functions
@njit
def any_peak( curve, prominence_threshold ):
    # numba-optimized function to check if there is a peak
    # with the given prominence in the data
    i_peak = np.argmax( curve )
    left_min = np.min(curve[:i_peak]) if i_peak > 0 else 0
    right_min = np.min(curve[i_peak + 1:]) if i_peak < curve.size-1 else 0
    prominence = curve[i_peak] - max(left_min, right_min)
    if prominence >= prominence_threshold:
        return True
    return False

@njit(parallel=True)
def _regroup_q_custom_baseline( projection, t_mask, mask_detector, bl_values, q_mask_p ):
    n_data = t_mask.shape[0] # effective number of images
    n_peaks = q_mask_p.shape[0]
    regrouped_data = np.empty( (n_data, projection.shape[1], n_peaks), data_type )

    for k in prange(n_data):
        t = t_mask[k] # index within projection
        # get image and subtract baseline
        data_k = projection[t]

        # regroup data into peaks
        data_k_regrouped = np.empty( (data_k.shape[0], n_peaks), data_type )
        for p in range(n_peaks):
            #calculate integrated baseline:
            base_p = bl_values[t,p]
            # subtract from peak-integral
            data_k_regrouped[:,p] = nb.masked_sum_axis1(data_k[:,q_mask_p[p]], mask_detector[:,q_mask_p[p]]) - base_p

        regrouped_data[k] = data_k_regrouped
    return regrouped_data

@njit(parallel=True)
def _regroup_q_baseline( projection, t_mask, mask_detector, bl_fun, bl_args, q_in, q_mask_bl, q_mask_p ):
    n_data = t_mask.shape[0] # effective number of images
    n_peaks = q_mask_p.shape[0]
    regrouped_data = np.empty( (n_data, projection.shape[1], n_peaks), data_type )

    for k in prange(n_data):
        t = t_mask[k] # index within projection
        # get image and subtract baseline
        data_k = projection[t]
        # data_k_1d = nb.nb_mean_ax0( projection[t] )
        data_k_1d = nb.masked_mean_axis0( projection[t], mask_detector )
        baseline = bl_fun( q_in, data_k_1d, q_mask_bl, bl_args )

        # regroup data into peaks
        data_k_regrouped = np.empty( (data_k.shape[0], n_peaks), data_type )
        for p in range(n_peaks):
            #calculate integrated baseline:
            base_p = baseline[q_mask_p[p]].sum()
            # subtract from peak-integral
            data_k_regrouped[:,p] = nb.masked_sum_axis1(data_k[:,q_mask_p[p]], mask_detector[:,q_mask_p[p]]) - base_p

        regrouped_data[k] = data_k_regrouped#.flatten()[mask_detector] # Attention! change! -> masking moved to fitting
    return regrouped_data

@njit(parallel=True)
def _regroup_q_azimuthally_resolved_baseline( projection, t_mask, mask_detector, bl_fun, bl_args, q_in, q_mask_bl, q_mask_p ):
    n_data = t_mask.shape[0] # effective number of images
    n_chi = projection.shape[1]
    n_peaks = q_mask_p.shape[0]
    regrouped_data = np.empty( (n_data, n_chi, n_peaks), data_type )

    for k in prange(n_data):
        t = t_mask[k] # index within projection
        # get image and subtract baseline
        data_k = projection[t]
        # data_k_1d = nb.nb_mean_ax0( projection[t] )

        # regroup data into peaks
        data_k_regrouped = np.empty( (data_k.shape[0], n_peaks), data_type )
        for p in range(n_peaks):
            data_kp = nb.masked_sum_axis1(data_k[:,q_mask_p[p]], mask_detector[:,q_mask_p[p]])
            for c in range(n_chi):
                data_k_chi = projection[t,c][mask_detector[c]]
                baseline = bl_fun( q_in, data_k_chi, q_mask_bl, bl_args )
                #calculate integrated baseline:
                base_p = baseline[q_mask_p[p]].sum()
                # subtract from peak-integral
                data_k_regrouped[c,p] = data_kp[c] - base_p

        regrouped_data[k] = data_k_regrouped#.flatten()[mask_detector] # Attention! change! -> masking moved to fitting
    return regrouped_data

@njit(parallel=True)
def _rescale( data, norm ):
    for k in prange( data.shape[0] ):
        if norm[k] == 0:
            data[k] = 0
        else:
            data[k] /= norm[k]
    return data

def import_data_1d( path, pattern, mod:model_textom,
                geo_path='input/geometry.py',
                flip_fov=False, use_ion=True ):

    geo = import_module_from_path('geometry', geo_path)
    # get the images that show the sample
    scanmask = mod.Beams[:,:,0].astype(bool)                           ###########

    print('Starting data import')
    print('\tLoading integrated data from files')
    t0=time()
    airscat, ion = [], []
    filelist = sorted( os.listdir( os.path.join(path,'data_integrated_1d')) )
    filelist = [s for s in filelist if pattern in s]#[:2]
    t1=time()
    for g, file in enumerate( filelist ):
        # Read data from integrated file
        with h5py.File(os.path.join(path, 'data_integrated_1d', file),'r') as hf:
            q_in = ( hf['radial_units'][0] ).astype(data_type)
            fov = ( hf['fov'][()] ).astype(np.int32)
            d = np.array(hf['cake_integ'])[()]
            if 'ion' in hf:
                ion.append(hf['ion'][()])

        if flip_fov:
            fov = np.flip( fov )
            
        if ion and use_ion:
            # print('\tRescale by beam intensity')
            d = _rescale( d, ion[g] )

        ## Reshaping in function of scanning mode ###
        # Base code written for column scan
        d = d.reshape( *fov, d.shape[1] )
        
        if 'line' in geo.scan_mode:
            fov = np.flip( fov )
            d = np.transpose( d, axes=(1,0,2)) 

        # reorder data so that they are all treated the same in the model
        if 'snake' in geo.scan_mode:
            # Flip every second row
            for ii in range(d.shape[0]):
                if ii % 2 != 0:
                    d[ii] = d[ii][::-1]
        ##############################################

        i0=d.shape[-1]//4
        # get airscattering:
        edge_sample = np.array([ 
            d[0,0,i0:].mean(), 
            d[0,-1,:,i0:].mean(), 
            d[0,-fov[1]//2,:,i0:].mean(),  
            d[fov[0]//2,0,i0:].mean(), 
            d[fov[0]//2,-1,:,i0:].mean(), 
            d[-1,0,:,i0:].mean(), 
            d[-1,-fov[1]//2,:,i0:].mean(),
            d[-1,-1,:,i0:].mean()
            ]) 
        airscat.append( np.min( edge_sample[edge_sample > 0.] ))
        if not ion or not use_ion:
            # print('\tRescale by air scattering')
            d /= airscat[g]

        max_shape = mod.fov
        # pad the data as in mumottize
        # fill projections with air scattering
        proj =  np.random.normal(airscat[g],np.sqrt(airscat[g])/2,(*max_shape, d.shape[2]))
        si, sj = d.shape[:2]
        i0 = (max_shape[0] - si)//2
        j0 = (max_shape[1] - sj)//2
        proj[i0:i0+si, j0:j0+sj] = d
        proj = proj.reshape(max_shape[0]*max_shape[1], *q_in.shape)

        scanmask_g = scanmask[g].copy()
        proj = _remove_crazypixels_1d(proj, scanmask_g)

        out_path = os.path.join( path, 'analysis', 'data_1drec.h5')
        if g == 0:
            print('\tSaving data to file: %s' % out_path)
            with h5py.File( out_path, 'w') as hf:
                hf.create_dataset( 'data',
                        shape=(0, proj.shape[1]),
                        maxshape=(None, proj.shape[1]),
                        chunks=(1, proj.shape[1]),
                        dtype='float64'
                    )
                hf.create_dataset( 'radial_units', data = q_in )

        # add data to h5 file
        with h5py.File( out_path, 'r+') as hf:
            dset = hf['data']
            current_rows = dset.shape[0]
            new_rows = current_rows + proj.shape[0]
            dset.resize( new_rows, axis=0)
            dset[current_rows:new_rows, :] = proj


        t_it = (time()-t1)/(g+1)
        Nrot = len(filelist)
        sys.stdout.write(f'\r\t\tProjection {(g+1):d} / {Nrot:d}, t/proj: {t_it:.1f} s, t left: {((Nrot-g-1)*t_it/60):.1f} min' )
        sys.stdout.flush()

    ion_av = np.array( [i.mean() for i in ion] ) # cannon write ion directly because shape
    gt_mask = np.array( np.where(scanmask) ).T
    # add metadata to h5 file
    with h5py.File( out_path, 'r+') as hf:
        hf.create_dataset( 'scanmask', data=scanmask )
        hf.create_dataset( 'gt_mask', data=gt_mask )
        hf.create_dataset( 'airscat', data = airscat )
        hf.create_dataset( 'ion', data = ion_av )
        hf.create_dataset( 'q', data=q_in)
        hf.create_dataset( 'textom_version', data=__version__)

    print(f'\t\ttook {time()-t0:.2f} s')

@njit(parallel=True)
def _remove_crazypixels_1d( projection, scanmask ):
    dat_new = projection.copy()
    for k in prange(projection.shape[0]):
        if scanmask[k]:
            for m in range(-1,projection.shape[1]-1):
                sample = np.array([
                    projection[k,m-2],projection[k,m-1],projection[k,m+1],projection[k,m+2]
                ])
                base = np.median(sample)
                if projection[k,m] > 10*np.abs(base):
                    dat_new[k,m] = base
    return dat_new


import matplotlib.pyplot as plt
import numpy as np
import os

import textom.src.handle as hdl
from textom.src.misc import import_module_from_path

def geometry_plot( geo ):

    # Create a figure and axis with 3D projection
    fig = plt.figure(figsize=(10, 8))
    ax = fig.add_subplot(111, projection='3d')

    # Define the coordinates for the 3D plot with offset
    X = np.array([-100, 100])  # x coordinates for the red line
    Y = np.array([0, 0])    # y coordinates for the red line
    Z = np.array([0, 0])    # z coordinates for the red line

    # Plot the red line in x-direction
    ax.plot(X, Y, Z, color='red')

    # Move the small coordinate system by -25 in y and z
    offset = -25
    ax.quiver(0, offset, offset, 50, 0, 0, color='b', label='X axis')  # X axis
    ax.quiver(0, offset, offset, 0, 50, 0, color='g', label='Y axis')  # Y axis
    ax.quiver(0, offset, offset, 0, 0, 50, color='purple', label='Z axis')  # Z axis

    # Generate random data with Gaussian distribution around a circle of radius 30 in y-z plane
    np.random.seed(0)  # for reproducibility
    n=100
    theta = np.random.uniform(0, 2*np.pi, n)
    r0 = np.random.normal(0, 5, n)  
    r1 = np.random.normal(15, 5, n) 
    r2 = np.random.normal(30, 2, n)  
    ax.scatter([100] * n, r0 * np.cos(theta), r0 * np.sin(theta), c='b', marker='.')
    ax.scatter([100] * n, r1 * np.cos(theta), r1 * np.sin(theta), c='b', marker='.')
    ax.scatter([100] * n, r2 * np.cos(theta), r2 * np.sin(theta), c='b', marker='.')

    # Add a small cube at the origin using a surface plot
    ax.scatter(0, 0, 0, c='y', marker='D', s=100, label='sample')

    # Add a curved arrow for rotation in the xy plane next to the sample
    arrow_radius = 15
    theta_curve = np.linspace(0, np.pi, 100)
    x_curve = arrow_radius * np.cos(theta_curve)
    y_curve = arrow_radius * np.sin(theta_curve)
    z_curve = np.zeros_like(x_curve)
    ax.plot(x_curve, y_curve, z_curve, color='g', linewidth=2)
    ax.quiver(arrow_radius,0,0,0,-5,0, color='g', linewidth=2, 
            length=1, arrow_length_ratio=2, label='inner rot axis')

    # Add a curved arrow for rotation in the xy plane next to the sample
    arrow_radius = 20
    theta_curve = np.linspace(0, np.pi, 100)
    x_curve = arrow_radius * np.cos(theta_curve)
    z_curve = arrow_radius * np.sin(theta_curve)
    y_curve = np.zeros_like(x_curve)
    ax.plot(x_curve, y_curve, z_curve, color='g', linewidth=2)
    ax.quiver(arrow_radius,0,0,0,0,-5, color='g', linewidth=2, 
            length=1, arrow_length_ratio=2, label='outer rot axis')

    # Set plot labels and limits
    ax.set_xlim([-100, 150])
    ax.set_ylim([-50, 50])
    ax.set_zlim([-50, 50])

    # Add a legend
    ax.legend()

    # Turn off the grid
    ax.grid(False)

    # Hide axes
    ax.set_axis_off()

    ax.set_aspect("equal")

    plt.show()

geo_path = hdl.get_file_path('textom',os.path.join('input','geometry.py'))
geo = import_module_from_path('geometry', geo_path)



import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.mplot3d import Axes3D

def plot_unit_sphere_with_points(angles, angles2):
    """
    Plots a 3D unit sphere and points on its surface given by a list of azimuthal and polar angles.

    Parameters:
    angles (list of tuples): List of (azimuthal, polar) angle tuples in radians.
    """
    # Create a figure
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    # Create a unit sphere
    u = np.linspace(0, 2 * np.pi, 100)
    v = np.linspace(0, np.pi, 100)
    x = np.outer(np.cos(u), np.sin(v))
    y = np.outer(np.sin(u), np.sin(v))
    z = np.outer(np.ones(np.size(u)), np.cos(v))
    ax.plot_surface(x, y, z, color='b', alpha=0.1)

    # # Create a band up to 45 deg tilt
    # u = np.linspace(0, 2 * np.pi, 100)
    # v = np.linspace(np.pi/4, np.pi/2, 100)
    # x = np.outer(np.cos(u), np.sin(v))
    # y = np.outer(np.sin(u), np.sin(v))
    # z = np.outer(np.ones(np.size(u)), np.cos(v))
    # ax.plot_surface(x, y, z, color='b', alpha=0.2)

    # Add latitude and longitude lines (grid)
    for theta in np.linspace(0, 2 * np.pi, 24):  # Longitude lines
        x = np.cos(theta) * np.sin(v)
        y = np.sin(theta) * np.sin(v)
        z = np.cos(v)
        ax.plot(x, y, z, color='k', linewidth=0.2)

    for phi in np.linspace(0, np.pi, 12):  # Latitude lines
        x = np.cos(u) * np.sin(phi)
        y = np.sin(u) * np.sin(phi)
        z = np.ones_like(u) * np.cos(phi)
        ax.plot(x, y, z, color='k', linewidth=0.2)

    # Plot points on the sphere
    for az, pol in angles:
        x_point = np.cos(az) * np.sin(pol)
        y_point = np.sin(az) * np.sin(pol)
        z_point = np.cos(pol)
        ax.scatter(x_point, y_point, z_point, color='r')

    # Plot points on the sphere
    for az, pol in angles2:
        x_point = np.cos(az) * np.sin(pol)
        y_point = np.sin(az) * np.sin(pol)
        z_point = np.cos(pol)
        ax.scatter(x_point, y_point, z_point, color='k')

    # Set labels
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')

    # Set the aspect ratio to be equal
    ax.set_box_aspect([1, 1, 1])  # Aspect ratio is 1:1:1 (equal)

    # Make grid and axes invisible
    ax.grid(False)
    # ax.set_axis_off()
    ax.set_xticks([])
    ax.set_xticks([], minor=True)
    ax.set_yticks([])
    ax.set_yticks([], minor=True)    
    ax.set_zticks([])
    ax.set_zticks([], minor=True)

    # d = 1.4
    # # Plot custom axes through the origin
    # ax.plot([-d, d], [0, 0], [0, 0], color='k', linewidth=1)
    # ax.plot([0, 0], [-d, d], [0, 0], color='k', linewidth=1)
    # ax.plot([0, 0], [0, 0], [-d, d], color='k', linewidth=1)

    # # Add labels to the axes
    # ax.text(d, 0, 0, 'X', color='k', fontsize='large')
    # ax.text(0, d, 0, 'Y', color='k', fontsize='large')
    # ax.text(0, 0, d, 'Z', color='k', fontsize='large')

    # Set the view to see the xz-plane
    ax.view_init(elev=0, azim=90)

    plt.show()

import src.handle as hdl
kappa, omega, shifty, shiftz, _, _ = hdl.load_mat_proj('data/SASTT_helix_s7_20230921_aligned_FBP.mat')

for k in range(kappa.shape[0]):
    if omega[k] > 180:
        kappa[k] *= -1
        omega[k] += 180

angles = np.column_stack((omega, 90-kappa))*np.pi/180

n=41
Ng0 = kappa.size
c=np.arange(0,Ng0,int(Ng0/n))[:n]
angles2 = angles[c]

plot_unit_sphere_with_points(angles, angles2)
